import argparse
import configparser
import enum
import re
import warnings
from dataclasses import asdict, dataclass, field
from typing import List

import json5


@dataclass
class Parameters:
    """
    Parameters class: a dataclass which contains all parameters required for initialising a pygwb Interferometer, a pygwb Baseline, and run pygwb_pipe.

    Attributes
    ----------
    t0 : float
        Initial time.
    tf: float
        Final time.
    interferometer_list: list
        List of interferometers to run the analysis with. Default is [\"H1\", \"L1\"]
    data_type: str
        Type of data to access/download; options are private, public, local. Default is public.
    channel: str
        Channel name; needs to match an existing channel. Default is \"GWOSC-16KHZ_R1_STRAIN\"
    frametype: str
        Frame type; Optional, desired channel needs to be found in listed frametype. Only used when data_type=private. Default is empty.
    new_sample_rate: int
        Sample rate to use when downsampling the data (Hz). Default is 4096 Hz.
    input_sample_rate: int
        Sample rate of the read data (Hz). Default is 16384 Hz.
    cutoff_frequency: int
        Lower frequency cutoff; applied in filtering in preprocessing (Hz). Default is 11 Hz.
    segment_duration: int
        Duration of the individual segments to analyse (seconds). Default is 192 seconds.
    number_cropped_seconds: int
        Number of seconds to crop at the start and end of the analysed data (seconds). Default is 2 seconds.
    window_downsampling: str
        Type of window to use in preprocessing. Default is \"hamming\"
    ftype: str
        Type of filter to use in downsampling. Default is \"fir\"
    frequency_resolution: float
        Frequency resolution of the final output spectrum (Hz). Default is 1\/32 Hz.
    polarization: str
        Polarisation type for the overlap reduction function calculation; options are scalar, vector, tensor. Default is tensor.
    alpha: float
        Spectral index to filter the data for. Default is 0.
    fref: int
        Reference frequency to filter the data at (Hz). Default is 25 Hz.
    flow: int
        Lower frequency to include in the analysis (Hz). Default is 20 Hz.
    fhigh: int
        Higher frequency to include in the analysis (Hz). Default is 1726 Hz.
    local_data_path: str
        Path(s) to local data, if the local data option is chosen. Default is empty.
    notch_list_path: str
        Path to the notch list file. Default is empty.
    coarse_grain_psd: bool
        Whether to apply coarse graining to obtain PSD spectra. Default is False.
    coarse_grain_csd: bool
        Whether to apply coarse graining to obtain CSD spectra. Default is True.
    overlap_factor_welch: float
        Overlap factor to use when if using Welch's method to estimate spectra (NOT coarsegraining). For \"hann\" window use 0.5 overlap_factor and for \"boxcar\" window use 0 overlap_factor. Default is 0.5 (50% overlap), which is optimal when using Welch's method with a \"hann\" window.
        Whether to apply coarse graining to obtain CSD spectra. Default is True.
    N_average_segments_psd: int
        Number of segments to average over when calculating the psd with Welch method. Default is 2.
    window_fft_dict: dict
        Dictionary containing name and parameters describing which window to use when producing fftgrams for psds and csds. Default is \"hann\".
    window_fft_dict_welch: dict 
        Dictionary containing name and parameters relative to which window to use when producing fftgrams for pwelch calculation. Default is "hann".
    calibration_epsilon: float
        Calibation coefficient. Default is 0.
    overlap_factor: float
        Factor by which to overlap consecutive segments for analysis. Default is 0.5 (50%% overlap)
    delta_sigma_cut: float
        Cutoff value for the delta sigma cut. Default is 0.2.
    alphas_delta_sigma_cut: list
        List of spectral indexes to use in delta sigma cut calculation. Default is [-5, 0, 3].
    save_data_type: str
        Suffix for the output data file. Options are hdf5, npz, json, pickle. Default is json.
    time_shift: int
        Seconds to timeshift the data by in preprocessing. Default is 0.
    path_gate_data: str
        Path to the map with pygwb output containing information about gates. If loading a single file, it has to be an .npzfile 
        with the same structure as a pygwb output file. Default is an empty string.
    gate_data: bool
        Whether to apply self-gating to the data in preprocessing. Default is False.
    gate_tzero: float
        Gate tzero. Default is 1.0.
    gate_tpad: float
        Gate tpad. Default is 0.5.
    gate_threshold: float
        Gate threshold. Default is 50.
    cluster_window: float
        Cluster window. Default is 0.5.
    gate_whiten: bool
        Whether to whiten when gating. Default is True.
    tag: str
        Hint for the read_data function to retrieve one specific type of data, e.g.: C00, C01
    return_naive_and_averaged_sigmas: bool
        option to return naive and sliding sigmas from delta sigma cut
    window_fftgram_dict: dictionary, optional
        Dictionary with window characteristics. Default is `(window_fftgram_dict={"window_fftgram": "hann"}`
    """

    t0: float = 0
    tf: float = 100
    interferometer_list: List = field(default_factory=lambda: ["H1", "L1"])
    data_type: str = "public"
    channel: str = "GWOSC-16KHZ_R1_STRAIN"
    frametype: str = ""
    new_sample_rate: int = 4096
    input_sample_rate: int = 16384
    cutoff_frequency: float = 11
    segment_duration: int = 192
    number_cropped_seconds: int = 2
    window_downsampling: str = "hamming"
    ftype: str = "fir"
    frequency_resolution: float = 0.03125
    polarization: str = "tensor"
    alpha: float = 0
    fref: float = 25
    flow: float = 20
    fhigh: float = 1726
    local_data_path: str = ""
    notch_list_path: str = ""
    coarse_grain_psd: bool = False
    coarse_grain_csd: bool = True
    overlap_factor_welch: float = 0.5
    N_average_segments_psd: int = 2
    window_fft_dict: dict = field(default_factory=lambda: {"window_fftgram": "hann"})
    window_fft_dict_welch: dict = field(default_factory=lambda: {"window_fftgram": "hann"})
    calibration_epsilon: float = 0
    overlap_factor: float = 0.5
    delta_sigma_cut: float = 0.2
    alphas_delta_sigma_cut: List = field(default_factory=lambda: [-5, 0, 3])
    save_data_type: str = "npz"
    time_shift: int = 0
    path_gate_data: str = ""
    gate_data: bool = False
    gate_tzero: float = 1.0
    gate_tpad: float = 0.5
    gate_threshold: float = 50.0
    cluster_window: float = 0.5
    gate_whiten: bool = True
    tag: str = ""
    return_naive_and_averaged_sigmas: bool = False


    def update_from_dictionary(self, kwargs):
        """Update parameters from a dictionary

        Parameters
        ----------
        kwargs: dictionary
            Dictionary of parameters to update.
        """
        ann = getattr(self, "__annotations__", {})
        for name, dtype in ann.items():
            if name in kwargs:
                try:
                    if not bool(re.search("^.+:.+,.+:.+$", kwargs[name])): 
                        kwargs[name] = dtype(kwargs[name]) if kwargs[name] != 'False' else False
                except TypeError:
                    pass
                setattr(self, name, kwargs[name])
        for name in kwargs:
            if name not in ann.keys():
                warnings.warn(
                    f"{name} is not an expected parameter and will be ignored."
                )

    def update_from_file(self, path: str) -> None:
        """Update parameters from an ini file

        Parameters
        ----------
        path: str
            Path to parameters ini file to use to update class.
        """
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(path)
        mega_list = config.items("data_specs")
        mega_list.extend(config.items("preprocessing"))
        mega_list.extend(config.items("density_estimation"))
        mega_list.extend(config.items("postprocessing"))
        mega_list.extend(config.items("gating"))
        mega_list.extend(config.items("data_quality"))
        mega_list.extend(config.items("output"))
        mega_list.extend(config.items("local_data"))
        dictionary = dict(mega_list)
        if dictionary["alphas_delta_sigma_cut"]:
            dictionary["alphas_delta_sigma_cut"] = json5.loads(
                dictionary["alphas_delta_sigma_cut"]
            )
        if dictionary["interferometer_list"]:
            dictionary["interferometer_list"] = json5.loads(
                dictionary["interferometer_list"]
            )
        dictionary["window_fft_dict"] = dict(config.items("window_fft_specs"))
        dictionary["window_fft_dict_welch"] = dict(config.items("window_fft_welch_specs"))
        for item in dictionary.copy():
            if not dictionary[item]:
                dictionary.pop(item)
        self.update_from_dictionary(dictionary)

    def update_from_arguments(self, args: List[str]) -> None:
        """Update parameters from a set of arguments

        Parameters
        ----------
        args: list
            List of arguments to update in the Class. Format must coincide to argparse formatting, e.g.,
            ['--t0', '0', '--tf', '100']

        Notes
        -----
        Not all possible options are available through argument updating. The two exceptions are the dictionary
        attributes which can not be parsed easily by argparse. These are
        * window_fft_dict: this is composed by the single argument
            -- window_fftgram
        This is the only option currently supported. To use windows that require extra parameters, pass these as part of an
        .ini file, in the relevant section [window_fft_specs].
        * window_fft_dict_welch: this is composed by the single argument
            -- window_fftgram_welch
        """
        if not args:
            return
        ann = getattr(self, "__annotations__", {})
        parser = argparse.ArgumentParser()
        for name, dtype in ann.items():
            if dtype == List:
                parser.add_argument(f"--{name}", type=str, nargs='+', required=False)
            else:
                parser.add_argument(f"--{name}", type=str, required=False)

        parser.add_argument("--window_fftgram", type=str, required=False)
        parser.add_argument("--window_fftgram_welch", type=str, required=False)
        parsed, _ = parser.parse_known_args(args)
        dictionary = vars(parsed)
        for item in dictionary.copy():
            if dictionary[item] is None:
                dictionary.pop(item)
        if "window_fftgram" in dictionary:
            window_fft_dict = {}
            wfgram_name, *wfgram = dictionary.pop("window_fftgram").split(',')
            window_fft_dict["window_fftgram"] = wfgram_name
            if wfgram_name.lower()=='tukey':
                window_fft_dict["alpha"] = wfgram[0]
            elif wfgram_name.lower()=='hann':
                pass
            else: 
                raise ValueError(f"Window {wfgram_name} not supported from command line! Please try submitting through a full parameter ini file.")
            dictionary["window_fft_dict"] = window_fft_dict
        if "window_fftgram_welch" in dictionary:
            window_fft_dict_welch = {}
            wfgram_name_welch, *wfgram_welch = dictionary.pop("window_fftgram_welch").split(',')
            window_fft_dict_welch["window_fftgram"] = wfgram_name_welch
            if wfgram_name_welch.lower()=='tukey':
                window_fft_dict_welch["alpha"] = wfgram[0]
            elif wfgram_name_welch.lower()=='hann':
                pass
            else: 
                raise ValueError(f"Window {wfgram_name_welch} not supported from command line! Please try submitting through a full parameter ini file.")
            dictionary["window_fft_dict_welch"] = window_fft_dict_welch
        self.update_from_dictionary(dictionary)

    def save_paramfile(self, output_path):
        """Save parameters to a parameters ini file.

        Parameters
        ----------
        output_path: str
            Full path for output parameters ini file.
        """
        param = configparser.ConfigParser()
        param.optionxform = str
        param_dict = asdict(self)
        # for key, value in param_dict.items():
        #    param_dict[key] = str(value)
        data_specs_dict = {}
        data_specs_dict["t0"] = param_dict["t0"]
        data_specs_dict["tf"] = param_dict["tf"]
        data_specs_dict["interferometer_list"] = param_dict["interferometer_list"]
        data_specs_dict["data_type"] = param_dict["data_type"]
        data_specs_dict["channel"] = param_dict["channel"]
        data_specs_dict["frametype"] = param_dict["frametype"]
        data_specs_dict["time_shift"] = param_dict["time_shift"]
        param["data_specs"] = data_specs_dict

        preprocessing_dict = {}
        preprocessing_dict["new_sample_rate"] = param_dict["new_sample_rate"]
        preprocessing_dict["cutoff_frequency"] = param_dict["cutoff_frequency"]
        preprocessing_dict["segment_duration"] = param_dict["segment_duration"]
        preprocessing_dict["number_cropped_seconds"] = param_dict[
            "number_cropped_seconds"
        ]
        preprocessing_dict["window_downsampling"] = param_dict["window_downsampling"]
        preprocessing_dict["ftype"] = param_dict["ftype"]
        param["preprocessing"] = preprocessing_dict
        
        gating_dict = {}
        gating_dict["path_gate_data"] = param_dict["path_gate_data"]
        gating_dict["gate_data"] = param_dict["gate_data"]
        gating_dict["gate_whiten"] = param_dict["gate_whiten"]
        gating_dict["gate_tzero"] = param_dict["gate_tzero"]
        gating_dict["gate_tpad"] = param_dict["gate_tpad"]
        gating_dict["gate_threshold"] = param_dict["gate_threshold"]
        gating_dict["cluster_window"] = param_dict["cluster_window"]
        param["gating"] = gating_dict

        param["window_fft_specs"] = self.window_fft_dict
        param["window_fft_welch_specs"] = self.window_fft_dict_welch

        density_estimation_dict = {}
        density_estimation_dict["frequency_resolution"] = param_dict[
            "frequency_resolution"
        ]
        density_estimation_dict["N_average_segments_psd"] = param_dict[
            "N_average_segments_psd"
        ]
        density_estimation_dict["coarse_grain_psd"] = param_dict["coarse_grain_psd"]
        density_estimation_dict["coarse_grain_csd"] = param_dict["coarse_grain_csd"]
        density_estimation_dict["overlap_factor_welch"] = param_dict["overlap_factor_welch"]
        density_estimation_dict["overlap_factor"] = param_dict["overlap_factor"]
        param["density_estimation"] = density_estimation_dict

        postprocessing_dict = {}
        postprocessing_dict["polarization"] = param_dict["polarization"]
        postprocessing_dict["alpha"] = param_dict["alpha"]
        postprocessing_dict["fref"] = param_dict["fref"]
        postprocessing_dict["flow"] = param_dict["flow"]
        postprocessing_dict["fhigh"] = param_dict["fhigh"]
        param["postprocessing"] = postprocessing_dict

        data_quality_dict = {}
        data_quality_dict["notch_list_path"] = param_dict["notch_list_path"]
        data_quality_dict["calibration_epsilon"] = param_dict["calibration_epsilon"]
        data_quality_dict["alphas_delta_sigma_cut"] = param_dict[
            "alphas_delta_sigma_cut"
        ]
        data_quality_dict["delta_sigma_cut"] = param_dict["delta_sigma_cut"]
        data_quality_dict["return_naive_and_averaged_sigmas"] = param_dict[
            "return_naive_and_averaged_sigmas"
        ]
        param["data_quality"] = data_quality_dict

        local_data_dict = {}
        local_data_dict["local_data_path"] = param_dict["local_data_path"]
        param["local_data"] = local_data_dict

        output_dict = {}
        output_dict["save_data_type"] = param_dict["save_data_type"]
        param["output"] = output_dict

        with open(output_path, "w") as configfile:
            param.write(configfile)

    def parse_ifo_parameters(self):
        ifo_parameters = ['channel', 'frametype', 'input_sample_rate', 'local_data_path', 'time_shift']
        ifo_list = self.interferometer_list
        param_dict = {}
        for ifo in ifo_list:
            param_dict[ifo] = Parameters()
        current_param_dict = self.__dict__
        for attr in current_param_dict.keys():
            if attr in ifo_parameters:
                attr_str = str(current_param_dict[attr])
                attr_split = attr_str.split(',')
                if len(attr_split)>1:
                    attr_dict = {key: value for key, value in (pair.split(':') for pair in attr_split)} 
                    for ifo in ifo_list:
                        param_dict[ifo].update_from_dictionary({attr: attr_dict[ifo]})
                else:
                    for ifo in ifo_list:
                        param_dict[ifo].update_from_dictionary({attr: current_param_dict[attr]})
            else:
                for ifo in ifo_list:
                    param_dict[ifo].update_from_dictionary({attr: current_param_dict[attr]})

        return param_dict


class ParametersHelp(enum.Enum):
    """Description of the arguments in the Parameters class. This is an enumeration class and is not meant for user interaction."""

    t0 = "Initial time."
    tf = "Final time."
    interferometer_list = (
        'List of interferometers to run the analysis with. Default is ["H1", "L1"]'
    )
    data_type = "Type of data to access/download; options are private, public, local. Default is public."
    channel = 'Channel name; needs to match an existing channel. Default is "GWOSC-16KHZ_R1_STRAIN" '
    frametype = 'Frame type; Optional, desired channel needs to be found in listed frametype. Only used when data_type=private. Default is empty. '
    new_sample_rate = (
        "Sample rate to use when downsampling the data (Hz). Default is 4096 Hz."
    )
    input_sample_rate = "Sample rate of the read data (Hz). Default is 16384 Hz."
    cutoff_frequency = "Lower frequency cutoff; applied in filtering in preprocessing (Hz). Default is 11 Hz."
    segment_duration = "Duration of the individual segments to analyse (seconds). Default is 192 seconds."
    number_cropped_seconds = "Number of seconds to crop at the start and end of the analysed data (seconds). Default is 2 seconds."
    window_downsampling = 'Type of window to use in preprocessing. Default is "hamming"'
    ftype = 'Type of filter to use in downsampling. Default is "fir"'
    frequency_resolution = (
        "Frequency resolution of the final output spectrum (Hz). Default is 1\/32 Hz."
    )
    polarization = "Polarisation type for the overlap reduction function calculation; options are scalar, vector, tensor. Default is tensor."
    alpha = "Spectral index to filter the data for. Default is 0."
    fref = "Reference frequency to filter the data at (Hz). Default is 25 Hz."
    flow = "Lower frequency to include in the analysis (Hz). Default is 20 Hz."
    fhigh = "Higher frequency to include in the analysis (Hz). Default is 1726 Hz."
    local_data_path = "Path(s) to local data, if the local data option is chosen. Default is empty."
    notch_list_path = "Path to the notch list file. Default is empty."
    coarse_grain_psd = "Whether to apply coarse graining to obtain PSD spectra. Default is False."
    coarse_grain_csd = "Whether to apply coarse graining to obtain CSD spectra. Default is True."
    overlap_factor_welch = "Overlap factor to use when if using Welch's method to estimate spectra (NOT coarsegraining). For \"hann\" window use 0.5 overlap_factor and for \"boxcar\" window use 0 overlap_factor. Default is 0.5 (50%% overlap), which is optimal when using Welch's method with a \"hann\" window."
    N_average_segments_psd = "Number of segments to average over when calculating the psd with Welch's method. Default is 2."
    window_fft_dict = 'Dictionary containing name and parameters relative to which window to use when producing fftgrams for psds and csds. Default is "hann".'
    window_fft_dict_welch = 'Dictionary containing name and parameters relative to which window to use when producing fftgrams for pwelch calculation. Default is "hann".'
    calibration_epsilon = "Calibation coefficient. Default is 0."
    overlap_factor = "Factor by which to overlap consecutive segments for analysis. Default is 0.5 (50%% overlap)"
    delta_sigma_cut = "Cutoff value for the delta sigma cut. Default is 0.2."
    alphas_delta_sigma_cut = "List of spectral indexes to use in delta sigma cut calculation. Default is [-5, 0, 3]."
    save_data_type = "Suffix for the output data file. Options are hdf5, npz, json, pickle. Default is json."
    time_shift = "Seconds to timeshift the data by in preprocessing. Default is 0."
    path_gate_data = (
        "Path to the pygwb output containing information about gates. If loading a single file, it has to be an .npzfile "
        "with the same structure as a pygwb output file. Default is an empty string."
    )
    gate_data = (
        "Whether to apply self-gating to the data in preprocessing. Default is False."
    )
    gate_tzero = "Gate tzero. Default is 1.0."
    gate_tpad = "Gate tpad. Default is 0.5."
    gate_threshold = "Gate threshold. Default is 50."
    cluster_window = "Cluster window. Default is 0.5."
    gate_whiten = "Whether to whiten when gating. Default is True."
    tag = "Hint for the read_data function to retrieve one specific type of data, e.g.: C00, C01"
    return_naive_and_averaged_sigmas = "option to return naive and sliding sigmas from delta sigma cut. Default value: False"

    @property
    def help(self):
        return self.value

    @property
    def argument(self):
        return self.name.lower()



import bilby
import gwpy
import numpy as np
from bilby.core.utils import create_frequency_series
from gwpy.timeseries import TimeSeries
from loguru import logger
from tqdm import tqdm

from pygwb.baseline import Baseline, get_baselines
from pygwb.util import interpolate_frequency_series


class Simulator(object):
    def __init__(
        self,
        interferometers,
        N_segments,
        duration,
        sampling_frequency,
        intensity_GW=None,
        injection_dict=None,
        start_time=0.0,
        no_noise=False,
        orf_polarization="tensor",
        seed=None,
        continuous=False,
    ):
        """
        Class that simulates an isotropic stochastic background.

        Parameters
        ==========
        interferometers: list of bilby interferometer objects
        intensity_GW: gwpy.frequencyseries.FrequencySeries
            A gwpy.frequencyseries.FrequencySeries containing the desired strain power spectrum
            which needs to be simulated. Note: A range of spectral indices (from -3 to 3) was
            tested. However, one should be careful for spectral indices outside of this range,
            as the splicing procedure implemented in this module is known to introduce a bias for
            some values of the spectral index (usually large negative numbers).
        injection_dict: dict
            Dictionary containing the parameters of CBC injections.
        N_segments: int
            Number of segments that needs to be generated for the simulation
        duration: float
            Duration of a simulated data segment
        sampling_frequency: float
            Sampling frequency
        start_time: float
            Start time of the simulation
        no_noise: boolean
            Flag that sets the noise_PSDs to 0
        seed: int
            Seed to use for random data generation
        continuous: boolean
            Flag to generate continuous segments. If True, a seed must also be provided 

        Returns
        =======
        """
        if len(interferometers) < 2:
            raise ValueError("Number of interferometers should be at least 2")
        else:
            self.interferometers = interferometers
            self.Nd = len(self.interferometers)

            self.sampling_frequency = sampling_frequency
            self.duration = duration

            for ifo in interferometers:
                ifo.sampling_frequency = self.sampling_frequency
                ifo.duration = self.duration

            self.N_segments = N_segments
            self.frequencies = self.get_frequencies()

            self.Nf = len(self.frequencies)
            self.t0 = start_time
            self.N_samples_per_segment = int(self.sampling_frequency * self.duration)
            self.deltaT = 1 / self.sampling_frequency

            self.noise_PSD_array = self.get_noise_PSD_array()

            self.no_noise = no_noise

            self.seed = seed
            self.continuous = continuous
            if (self.continuous == True) and not self.seed:
                raise ValueError("Must provide a seed to generate continuous segments")

            if no_noise == True:
                self.noise_PSD_array = np.zeros_like(self.noise_PSD_array)

            self.baselines = get_baselines(
                self.interferometers, frequencies=self.frequencies
            )
            for baseline in self.baselines:
                if not baseline._orf_polarization_set:
                    baseline.orf_polarization = "tensor"
            self.orf = self.get_orf(polarization=orf_polarization)

            self.intensity_GW = intensity_GW
            if self.intensity_GW:
                self.intensity_GW = interpolate_frequency_series(
                    intensity_GW, self.frequencies
                )
            self.injection_dict = injection_dict

            if not self.intensity_GW and not self.injection_dict:
                raise ValueError('Must provide at least one of intensity_GW or self.injection_dict')

    def get_frequencies(self):
        """
        Computes an array of frequencies with given sampling frequency and duration.

        Parameters
        ==========

        Returns
        =======
        frequencies: array_like
            Array containing the computed frequencies
        """
        frequencies = create_frequency_series(
            sampling_frequency=self.sampling_frequency, duration=self.duration
        )
        return frequencies

    def get_noise_PSD_array(self):
        """
        Function that gets the noise PSD array of all the interferometers.

        Parameters
        ==========

        Returns
        =======
        noise_PSDs_array: array_like
            Array containing the noise PSD arrays for all interferometers in
            self.interferometers.
        """
        noise_PSDs = []
        try:
            for ifo in self.interferometers:
                psd_temp = ifo.power_spectral_density.psd_array
                freqs_temp = ifo.power_spectral_density.frequency_array

                if np.isinf(psd_temp).any() == True:
                    raise ValueError(
                        f"The noisePSD of interferometer {ifo.name} contains infs!"
                    )
                psd = gwpy.frequencyseries.FrequencySeries(
                    psd_temp, frequencies=freqs_temp
                )
                psd_interpolated = interpolate_frequency_series(psd, self.frequencies)

                noise_PSDs.append(psd_interpolated.value[:])

                noise_PSDs_array = np.array(noise_PSDs)
            return noise_PSDs_array

        except:
            raise AttributeError(
                "The noisePSD of all the detectors needs to be specified!"
            )

    def get_orf(self, polarization="tensor"):
        """
        Function that returns a list containing the overlap reduction functions
        for all the baselines in self.baselines.

        Parameters
        ==========

        Returns
        =======
        orf_list: list
            List containing the overlap reduction functions for all the baselines
            in self.baselines.
        """
        orf_list = []
        for base in self.baselines:
            base.orf_polarization = polarization
        orfs = np.array(
            [baseline.overlap_reduction_function for baseline in self.baselines]
        )
        for orf in orfs:
            if orf.shape[0] != self.frequencies.shape[0]:
                orf = orf[1:]
            orf_list.append(orf)
        return orf_list

    def get_data_for_interferometers(self):
        """
        Get a data dictionary for interferometers

        Parameters
        ==========

        Returns
        =======
        interferometer_data: dict
            A dictionary with the simulated data for the interferometers.
        """
        data = self.generate_data(data_start=self.t0)
        interferometer_data = {}
        for idx, ifo in enumerate(self.interferometers):
            interferometer_data[ifo.name] = data[idx]
        if self.continuous:
            pad_len = int(self.N_segments * self.N_samples_per_segment/2)
            total_dur = self.N_segments*self.duration
            early_data = self.generate_data(data_start=self.t0-total_dur/2)
            late_data = self.generate_data(data_start=self.t0+total_dur/2)
            for idx, ifo in enumerate(self.interferometers):
                ed = early_data[idx] * TimeSeries(np.sin((early_data[idx].times.value-early_data[idx].t0.value
                                                   )/total_dur*np.pi), times=early_data[idx].times)
                ld = late_data[idx] * TimeSeries(np.sin((late_data[idx].times.value-late_data[idx].t0.value
                                                  )/total_dur*np.pi), times=late_data[idx].times)
                cd = interferometer_data[ifo.name] * TimeSeries(np.sin((interferometer_data[ifo.name].times.value-interferometer_data[ifo.name].t0.value
                                                            )/total_dur*np.pi), times=interferometer_data[ifo.name].times)
                interferometer_data[ifo.name] = \
                               ed.pad((0, 2*pad_len)) + \
                               cd.pad(pad_len) + \
                               ld.pad((2*pad_len, 0))
                interferometer_data[ifo.name] = \
                    interferometer_data[ifo.name].crop(self.t0,
                                              self.t0+total_dur)
                interferometer_data[ifo.name] = TimeSeries(interferometer_data[ifo.name].value, 
                                                           t0=self.t0, sample_rate=self.sampling_frequency)
        return interferometer_data

    def generate_data(self, data_start=None):
        """
        Function that simulates an isotropic stochastic background given the
        input parameters. The data is simulated and spliced together to prevent
        periodicity artifacts related to IFFTs.

        Parameters
        ==========
        data_start: float
            Start time of the data that is being generated.

        Returns
        =======
        data: array_like
            An array of size Nd (number of detectors) with gwpy TimeSeries with the
            data containing the simulated isotropic stochastic background.
        """
        if not data_start:
            data_start = self.t0
        data_signal_temp = np.zeros(
            (self.Nd, self.N_samples_per_segment * self.N_segments), dtype=np.ndarray
        )
        if self.intensity_GW:
            y_signal = self.simulate("signal")
            if self.seed:
                # seed is based on start time of segment
                np.random.seed(int(self.seed+(data_start%1000)))
            data_spliced = self.splice_segments(y_signal)
            for ii in range(self.Nd):
                data_signal_temp[ii] = data_signal_temp[ii] + data_spliced[ii]
        if self.injection_dict:
            data_cbc = self.inject_CBC()
            for ii in range(self.Nd):
                data_signal_temp[ii] = data_signal_temp[ii] + data_cbc[ii]
        if not self.no_noise:
            y_noise = self.simulate("noise")
            data_noise_temp = self.splice_segments(y_noise)
        else:
            data_noise_temp = np.zeros_like(data_signal_temp)
        data = np.zeros(self.Nd, dtype=gwpy.timeseries.TimeSeries)
        for ii in range(self.Nd):
            logger.info(
                f"Adding data to channel {self.interferometers[ii].name}:SIM-STOCH_INJ"
            )
            data[ii] = gwpy.timeseries.TimeSeries(
                (data_signal_temp[ii] + data_noise_temp[ii]).astype("float64"),
                t0=data_start,
                dt=self.deltaT,
                channel=f"{self.interferometers[ii].name}:SIM-STOCH_INJ",
                name=f"{self.interferometers[ii].name}:SIM-STOCH_INJ",
            )
        return data

    def orf_to_array(self):
        """
        Function that converts the list of overlap reduction functions into an array
        to facilitate the correct implementation when computing the covariance matrix.

        Parameters
        ==========

        Returns
        =======
        orf_array: array_like
            Array of shape Nd x Nd containing the orfs, where Nd is the number of
            detectors. The convention used for consistency with the remainder of the
            simulation is as follows. Orfs are only present in the off-diagonal slots
            in the array. Only the part below the diagonal is filled, after which this
            is copied to the upper part by transposing and summing. The array is filled
            by starting from the first free slot below the diagonal, from left to right,
            until the diagonal is reached, after which the line below in the array is
            filled analogously and so on.
        """
        index = 0
        orf_array = np.zeros(
            (self.Nd, self.Nd),
            dtype=gwpy.frequencyseries.FrequencySeries,
        )
        for ii in range(self.Nd):
            for jj in range(ii):
                orf_array[ii, jj] = self.orf[index]
                index += 1
        orf_array = orf_array + orf_array.transpose()

        for ii in range(self.Nd):
            baseline_name = (
                f"{self.interferometers[ii].name} - {self.interferometers[ii].name}"
            )
            baseline_temp = Baseline(
                baseline_name,
                self.interferometers[ii],
                self.interferometers[ii],
                frequencies=self.frequencies,
            )
            if not baseline_temp._orf_polarization_set:
                baseline_temp.orf_polarization = "tensor"
            orf_temp = baseline_temp.overlap_reduction_function

            if orf_temp.shape[0] != self.frequencies.shape[0]:
                orf_temp = orf_temp[1:]
            orf_array[ii, ii] = orf_temp

        return orf_array

    def covariance_matrix(self, flag):
        """
        Function to compute the covariance matrix corresponding to a stochastic
        background in the various detectors.

        Parameters
        ==========
        flaf: str
            Either flagged as "noise" or "signal", to allow for different generation of
            covariance matrix in both cases.

        Returns
        =======
        C: array_like
            Covariance matrix corresponding to a stochastic background in the
            various detectors. Dimensions are Nd x Nd x Nf, where Nd is the
            number of detectors and Nf the number of frequencies.
        """
        orf_array = self.orf_to_array()

        C = np.zeros((self.Nd, self.Nd, self.Nf))
        if flag == "noise":
            for ii in range(self.Nd):
                for jj in range(self.Nd):
                    if ii == jj:
                        C[ii, jj, :] = self.noise_PSD_array[ii]
        elif flag == "signal":
            for ii in range(self.Nd):
                for jj in range(self.Nd):
                    C[ii, jj, :] = orf_array[ii, jj] * self.intensity_GW.value[:]

        C[C == 0.0] = 1.0e-60

        C = self.N_samples_per_segment / (self.deltaT * 4) * C
        return C

    def compute_eigval_eigvec(self, C):
        """
        Function to compute the eigenvalues and eigenvectors of the covariance
        matrix corresponding to a stochastic background in the various detectors.

        Parameters
        ==========
        C: array_like
            Covariance matrix corresponding to a stochastic background in the
            various detectors. Dimensions are Nd x Nd x Nf, where Nd is the
            number of detectors and Nf the number of frequencies.

        Returns
        =======
        eigval: array_like
            Array of diagonal matrices containing the eigenvalues of the
            covariance matrix C.
        eigvec: array_like
            Array of matrices containing the eigenvectors of the covariance
            matrix C.
        """
        eigval, eigvec = np.linalg.eig(C.transpose((2, 0, 1)))
        eigval = np.array([np.diag(x) for x in eigval])
        return eigval, eigvec

    def generate_freq_domain_data(self):
        """
        Function that generates the uncorrelated frequency domain data with
        random phases for the stochastic background.

        Parameters
        ==========

        Returns
        =======
        z: array_like
            Array of size Nf x Nd containing uncorrelated frequency domain data.
        """
        z = np.zeros((self.Nf, self.Nd), dtype="complex_")
        re = np.random.randn(self.Nf, self.Nd)
        im = np.random.randn(self.Nf, self.Nd)
        z = re + im * 1j
        return z

    def transform_to_correlated_data(self, z, C):
        """
        Function that transforms the uncorrelated stochastic background
        simulated data, to correlated data.

        Parameters
        ==========
        z: array_like
            Array containing the uncorrelated data with random phase.
        C: array_like
            Array of size Nd x Nd x Nf representing the covariance matrices
            between detectors for a desired stochastic background, where Nd
            is the number of detectors and Nf is the number of frequencies.

        Returns
        =======
        x: array_like
            Array of size Nf x Nd, containing the correlated stochastic
            background data.
        """
        eigval, eigvec = self.compute_eigval_eigvec(C)

        A = np.einsum("...ij,jk...", np.sqrt(eigval + 0j), eigvec.transpose())
        x = np.einsum("...j,...jk", z, A)
        return x

    def simulate(self, flag):
        """
        Function that simulates the data corresponding to an isotropic stochastic
        background.

        Parameters
        ==========

        Returns
        =======
        y: array_like
            Array of size Nd x 2*(N_segments+1) x N_samples_per_segment containing the
            various segments with the simulated data.
        """
        C = self.covariance_matrix(flag)

        y = np.zeros(
            (self.Nd, 2 * self.N_segments + 1, self.N_samples_per_segment),
            dtype=np.ndarray,
        )

        for kk in range(2 * self.N_segments + 1):
            z = self.generate_freq_domain_data()

            xtemp = self.transform_to_correlated_data(z, C)

            for ii in range(self.Nd):
                if self.N_samples_per_segment % 2 == 0:
                    xtilde = np.concatenate(
                        (
                            np.array([0]),
                            xtemp[:, ii],
                            np.array([0]),
                            np.flipud(np.conjugate(xtemp[:, ii])),
                        )
                    )
                else:
                    xtilde = np.concatenate(
                        (
                            np.array([0]),
                            xtemp[:, ii],
                            np.flipud(np.conjugate(xtemp[:, ii])),
                        )
                    )
                y[ii, kk, :] = np.real(np.fft.ifft(xtilde))[
                    : self.N_samples_per_segment
                ]
        return y

    def splice_segments(self, segments):
        """
        This function splices together the various segments to prevent
        artifacts related to the periodicity that can arise from inverse
        Fourier transforms. Note: A range of spectral indices (from -3 to 3) was
        tested for the GW power spectrum to inject. However, one should be careful
        for spectral indices outside of this range, as the splicing procedure
        implemented here is known to introduce a bias for some values of the spectral
        index (usually large negative numbers).

        Parameters
        ==========
        segments: array_like
            Array of size Nd x (2*N_segments+1) x N_samples_per_segment containing the
            various segments with the simulated data that need to be spliced
            together, where Nd is the number of detectors.

        Returns
        =======
        data: array_like
            Array of size Nd x (N_segments*N_samples_per_segment) containing the simulated
            data corresponding to an isotropic stochastic background for each of the
            detectors, where Nd is the number of detectors.
        """
        w = np.zeros(self.N_samples_per_segment)

        for ii in range(self.N_samples_per_segment):
            w[ii] = np.sin(np.pi * ii / self.N_samples_per_segment)

        data = np.zeros(
            (self.Nd, self.N_samples_per_segment * self.N_segments), dtype=np.ndarray
        )

        for ii in range(self.Nd):
            for jj in range(self.N_segments):
                y0 = w * segments[ii][2 * jj][:]
                y1 = w * segments[ii][2 * jj + 1][:]
                y2 = w * segments[ii][2 * jj + 2][:]

                z0 = np.concatenate(
                    (
                        y0[
                            int(
                                self.N_samples_per_segment / 2
                            ) : self.N_samples_per_segment
                        ],
                        np.zeros(int(self.N_samples_per_segment / 2)),
                    )
                )
                z1 = y1[:]
                z2 = np.concatenate(
                    (
                        np.zeros(int(self.N_samples_per_segment / 2)),
                        y2[0 : int(self.N_samples_per_segment / 2)],
                    )
                )

                data[
                    ii,
                    jj
                    * self.N_samples_per_segment : (jj + 1)
                    * self.N_samples_per_segment,
                ] = (
                    z0 + z1 + z2
                )

        return data

    def inject_CBC(self):
        """
        This function uses the provided dictionary of injection parameters to 
        simulate a background of CBC sources. 

        Parameters
        ==========

        Returns
        =======
        data: array_like
            Array of size Nd x (N_segments*N_samples_per_segment) containing the simulated
            data corresponding to a background of CBC sources for each of the
            detectors, where Nd is the number of detectors.
        """

        data = np.zeros(
            (self.Nd, self.N_samples_per_segment * self.N_segments), dtype=np.ndarray
        )
        empty_ts = gwpy.timeseries.TimeSeries(np.zeros(self.N_samples_per_segment * self.N_segments),
                                              t0=self.t0, sample_rate = self.sampling_frequency)
        waveform_generator = bilby.gw.WaveformGenerator(
            duration=self.N_segments*self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
            waveform_arguments={
                "waveform_approximant": "IMRPhenomPv2",
                "reference_frequency": 50,
            },
        )
        for ii in range(self.Nd):
            det = bilby.gw.detector.get_empty_interferometer(self.interferometers[ii].name) 
            det.strain_data.set_from_gwpy_timeseries(empty_ts) 
            det.minimum_frequency=10.

            for n in tqdm(range(len(self.injection_dict['geocent_time']))):
                inj_params = {}
                for k in self.injection_dict.keys():
                    inj_params[k] = self.injection_dict[k][n]
            
                det.inject_signal(
                    waveform_generator=waveform_generator, parameters=inj_params
                )
            data[ii] = det.strain_data.to_gwpy_timeseries().value

        return data



import json
import warnings
from os import listdir
from os.path import isfile, join
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.transforms as mt
from astropy.time import Time
from loguru import logger

matplotlib.rcParams['figure.figsize'] = (8,6)
matplotlib.rcParams['axes.grid'] = True
matplotlib.rcParams['grid.linestyle'] = ':'
matplotlib.rcParams['grid.color'] = 'grey'
matplotlib.rcParams['lines.linewidth'] = 2
matplotlib.rcParams['legend.handlelength'] = 3

from matplotlib import rc

rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
# FIXME - changed to get workflow running
rc('text', usetex=False)

import seaborn as sns

sea = sns.color_palette("tab10")

import numpy as np
from scipy import integrate, stats
from scipy.optimize import curve_fit

from pygwb.baseline import Baseline
from pygwb.notch import StochNotchList
from pygwb.parameters import Parameters
from pygwb.util import StatKS, calc_bias


class StatisticalChecks(object):
    def __init__(
        self,
        sliding_times_all,
        sliding_omega_all,
        sliding_sigmas_all,
        naive_sigmas_all,
        coherence_spectrum,
        coherence_n_segs,
        point_estimate_spectrum,
        sigma_spectrum,
        frequencies,
        badGPSTimes,
        delta_sigmas,
        plot_dir,
        baseline_name,
        param_file,
        frequency_mask = None,
        gates_ifo1 = None,
        gates_ifo2 = None,
        file_tag = None,
        legend_fontsize = 16,
        convention = 'pygwb'
    ):
        """
        The statistical checks class performs various tests by plotting different quantities and saving this plots. This allows the user to check for consistency with expected results. Concretely, the following tests and plots can be generated: running point estimate, running sigma, (cumulative) point estimate integrand, real and imaginary part of point estimate integrand, FFT of the point estimate integrand, (cumulative) sensitivity, evolution of omega and sigma as a function of time, omega and sigma distribution, KS test, and a linear trend analysis of omega in time. Furthermore, part of these plots compares the values of these quantities before and after the delta sigma cut. Each of these plots can be made by calling the relevant class method (e.g. `plot_running_point_estimate()`).

        Parameters
        ==========
        sliding_times_all: array
            Array of GPS times before the bad GPS times from the delta sigma cut are applied.
        sliding_omega_all: array
            Array of sliding omegas before the bad GPS times from the delta sigma cut are applied.
        sliding_sigmas_all: array
            Array of sliding sigmas before the bad GPS times from the delta sigma cut are applied.
        naive_sigmas_all: array
            Array of naive sigmas before the bad GPS times from the delta sigma cut are applied.
        coherence_spectrum: array
            Array containing a coherence spectrum. Each entry in this array corresponds to the 2-detector coherence spectrum evaluated at the corresponding frequency in the frequencies array.
        coherence_n_segs: int
            Number of segments used for coherence calculation.
        point_estimate_spectrum: array
            Array containing the point estimate spectrum. Each entry in this array corresponds to the point estimate spectrum evaluated at the corresponding frequency in the frequencies array.
        sigma_spectrum: array
            Array containing the sigma spectrum. Each entry in this array corresponds to the sigma spectrum evaluated at the corresponding frequency in the frequencies array.
        frequencies: array
            Array containing the frequencies.
        badGPStimes: array
            Array of bad GPS times, i.e. times that do not pass the delta sigma cut.
        delta_sigmas: array
            Array containing the value of delta sigma for all times in sliding_times_all.
        plot_dir: str
            String with the path to which the output of the statistical checks (various plots) will be saved.
        baseline_name: str
            Name of the baseline under consideration.
        param_file: str
            String with path to the file containing the parameters that were used for the analysis run.
        frequency_mask: array
            Boolean mask applied to the specrtra in broad-band analyses. 
        gates_ifo1/gates_ifo2: list
            List of gates applied to interferometer 1/2.
        file_tag: str
            Tag to be used in file naming convention.
        legend_fontsize: int
            Font size for plot legends. Default is 16. All other fonts are scaled to this font.

        Returns
        =======
        Initializes an instance of the statistical checks class.
        """
        self.params = Parameters()
        self.params.update_from_file(param_file)

        self.sliding_times_all = sliding_times_all
        self.days_all = (sliding_times_all - sliding_times_all[0]) / 86400.0
        self.sliding_omega_all = sliding_omega_all
        self.sliding_sigmas_all = sliding_sigmas_all
        self.naive_sigmas_all = naive_sigmas_all
        self.badGPStimes = badGPSTimes
        self.delta_sigmas_all = delta_sigmas
        self.sliding_deviate_all = (
            self.sliding_omega_all - np.nanmean(self.sliding_omega_all)
        ) / self.sliding_sigmas_all
        self.gates_ifo1 = gates_ifo1
        self.gates_ifo2 = gates_ifo2

        self.frequencies = frequencies
        if frequency_mask is not None:
            self.frequency_mask = frequency_mask
        else:
            self.frequency_mask = True

        self.coherence_spectrum = coherence_spectrum
        fftlength = int(1.0 / (self.frequencies[1] - self.frequencies[0]))
        self.n_segs = coherence_n_segs*(1.-self.params.overlap_factor) * int(np.floor(self.params.segment_duration/(fftlength*(1.-self.params.overlap_factor_welch)))-1)
        if self.params.coarse_grain_csd:
            # Note: this breaks down when self.params.segment_duration/fftlength < 3
            self.n_segs = coherence_n_segs*(1.-self.params.overlap_factor) * int(np.floor(self.params.segment_duration/(fftlength)))
        self.n_segs_statement = r"The number of segments is" + f" {self.n_segs}."

        self.sigma_spectrum = sigma_spectrum
        self.point_estimate_spectrum = point_estimate_spectrum

        self.plot_dir = Path(plot_dir)

        self.baseline_name = baseline_name
        self.segment_duration = self.params.segment_duration
        self.deltaF = self.params.frequency_resolution
        self.new_sample_rate = self.params.new_sample_rate
        self.deltaT = 1.0 / self.new_sample_rate
        self.fref = self.params.fref
        self.flow = self.params.flow
        self.fhigh = self.params.fhigh

        self.alpha = self.params.alpha
        (
            self.sliding_times_cut,
            self.days_cut,
            self.sliding_omega_cut,
            self.sliding_sigma_cut,
            self.naive_sigma_cut,
            self.delta_sigmas_cut,
            self.sliding_deviate_cut,
            self.sliding_deviate_KS,
        ) = self.get_data_after_dsc()
        self.dsc_percent = (len(self.sliding_times_all) - len(self.sliding_times_cut))/len(self.sliding_times_all) * 100
        self.dsc_statement = r"The $\Delta\sigma$ cut removed" + f"{float(f'{self.dsc_percent:.2g}'):g}% of the data."

        tot_tot_segs = ((sliding_times_all - sliding_times_all[0])[-1])/(self.params.segment_duration*(1-self.params.overlap_factor))
        self.percent_obs_segs = len(self.naive_sigmas_all)/tot_tot_segs * 100

        (
            self.running_pt_estimate,
            self.running_sigmas,
        ) = self.compute_running_quantities()

        t0 = Time(self.sliding_times_all[0], format='gps')
        t0 = Time(t0, format='iso', scale='utc', precision=0, out_subfmt='date_hm')
        tf = Time(self.sliding_times_all[-1], format='gps')
        tf = Time(tf, format='iso', scale='utc', precision=0, out_subfmt='date_hm')
        self.time_tag = f"{t0}"+" $-$ "+f"{tf}"

        if file_tag:
            self.file_tag = file_tag
        else:
            self.file_tag = f"{self.sliding_times_all[0]}-{self.params.tf}"

        self.legend_fontsize = legend_fontsize
        self.axes_labelsize = legend_fontsize + 2
        self.title_fontsize = legend_fontsize + 4
        self.annotate_fontsize = legend_fontsize - 4

        ## convention: stochmon
        if convention == 'stochmon':
            self.days_all = self.days_all*24 + t0.ymdhms.hour + t0.ymdhms.minute/60
            self.days_cut = self.days_cut*24 + t0.ymdhms.hour + t0.ymdhms.minute/60
            self.xaxis = f"Hours since {t0}"
        else:
            self.xaxis = f"Days since {t0}"

    def get_data_after_dsc(self):
        """
        Function that returns the GPS times, the sliding omegas, the sliding sigmas, the naive sigmas, the delta sigmas and the sliding deviates after the bad GPS times from the delta sigma cut were applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_times_all`).

        Returns
        =======
        sliding_times_cut: array
            Array of GPS times after the bad GPS times were applied.
        days_cut: array
            Array of days after the bad GPS times were applied.
        sliding_omega_cut: array
            Array of the sliding omega values after the bad GPS times were applied.
        sliding_sigma_cut: array
            Array of sliding sigmas after the bad GPS times were applied.
        naive_sigma_cut: array
            Array of naive sigmas after the bad GPS times were applied.
        delta_sigma_cut: array
            Array of the delta sigma values after the bad GPS times were applied.
        sliding_deviate_cut: array
            Array of the deviates after the bad GPS times were applied.
        """
        bad_gps_times = self.badGPStimes
        bad_gps_mask = np.array([(t not in bad_gps_times) for t in self.sliding_times_all])

        bad_gps_mask[~np.isfinite(self.sliding_omega_all)] = False
        sliding_times_cut = self.sliding_times_all.copy()
        days_cut = self.days_all.copy()
        sliding_omega_cut = self.sliding_omega_all.copy()
        sliding_sigma_cut = self.sliding_sigmas_all.copy()
        naive_sigma_cut = self.naive_sigmas_all.copy()
        delta_sigma_cut = self.delta_sigmas_all.copy()
        sliding_deviate_cut = self.sliding_deviate_all.copy()

        sliding_times_cut = sliding_times_cut[bad_gps_mask]

        days_cut = days_cut[bad_gps_mask]
        sliding_omega_cut = sliding_omega_cut[bad_gps_mask]
        sliding_sigma_cut = sliding_sigma_cut[bad_gps_mask]
        naive_sigma_cut = naive_sigma_cut[bad_gps_mask]
        delta_sigma_cut = delta_sigma_cut[bad_gps_mask]
        sliding_deviate_cut = (
            sliding_omega_cut - np.nanmean(self.sliding_omega_all)
        ) / sliding_sigma_cut

        sliding_deviate_KS = (
            sliding_omega_cut - np.nanmean(sliding_omega_cut)
        ) / sliding_sigma_cut

        return (
            sliding_times_cut,
            days_cut,
            sliding_omega_cut,
            sliding_sigma_cut,
            naive_sigma_cut,
            delta_sigma_cut,
            sliding_deviate_cut,
            sliding_deviate_KS,
        )

    def compute_running_quantities(self):
        """
        Function that computes the running point estimate and running sigmas from the sliding point estimate and sliding sigmas. This is done only for the values after the delta sigma cut. This method does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_sigma_cut`).

        Returns
        =======
        running_pt_estimate: array
            Array containing the values of the running point estimate.
        running_sigmas: array
            Array containing the values of the running sigmas.
        """
        running_pt_estimate = self.sliding_omega_cut.copy()
        running_sigmas = self.sliding_sigma_cut.copy()

        ii = 0
        while ii < self.sliding_times_cut.shape[0] - 1:
            ii += 1
            numerator = np.nansum([running_pt_estimate[ii - 1] / (
                running_sigmas[ii - 1] ** 2
            ) , self.sliding_omega_cut[ii] / (self.sliding_sigma_cut[ii] ** 2)])
            denominator = np.nansum([1.0 / (running_sigmas[ii - 1] ** 2) , 1 / (
                self.sliding_sigma_cut[ii] ** 2
            )])
            running_pt_estimate[ii] = numerator / denominator
            running_sigmas[ii] = np.sqrt(1.0 / denominator)

        return running_pt_estimate, running_sigmas

    def compute_ifft_integrand(self):
        """
        Function that computes the inverse Fourier transform of the point estimate integrand. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_integrand`).

        Returns
        =======
        t_array: array
            Array containing the time lag values (in seconds).
        omega_t: array
            Array containing the

        """

        numFreqs = self.point_estimate_spectrum.shape[0]
        fhigh = self.flow + self.deltaF * numFreqs

        fNyq = 1 / (2 * self.deltaT)

        numFreqs_pre = np.floor(self.flow / self.deltaF) - 1
        f_pre = self.deltaF * np.arange(1, numFreqs_pre + 1)
        numFreqs_post = np.floor((fNyq - fhigh) / self.deltaF)
        f_post = fhigh + self.deltaF * np.arange(0, numFreqs_post)
        fp = np.concatenate((f_pre, self.frequencies, f_post))
        fn = -np.flipud(fp)
        f_tot = np.concatenate((fn, np.array([0]), fp))

        integrand_pre = np.zeros(int(numFreqs_pre))
        integrand_post = np.zeros(int(numFreqs_post))
        integrand_p = np.concatenate(
            (integrand_pre, 0.5 * self.point_estimate_spectrum, integrand_post)
        )

        integrand_n = np.flipud(np.conj(integrand_p))

        integrand_tot = np.concatenate((np.array([0]), integrand_p, integrand_n))

        fft_integrand = np.fft.fftshift(np.fft.fft(self.deltaF * integrand_tot))

        t_array = np.arange(
            -1.0 / (2 * self.deltaF) + self.deltaT, 1.0 / (2 * self.deltaF), self.deltaT
        )
        omega_t = np.flipud(fft_integrand)

        return t_array, omega_t

    def plot_running_point_estimate(self, ymin=None, ymax=None):
        """
        Generates and saves a plot of the running point estimate. The plotted values are the ones after the delta sigma cut. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `days_cut`).

        Parameters
        ==========
        ymin: float
            Minimum value on the y-axis.
        ymax: float
            Maximum value on the y-axis.

        """
        if self.days_cut.size==0:
            return
        fig = plt.figure(figsize=(10, 8))
        plt.plot(
            self.days_cut,
            self.running_pt_estimate,
            ".",
            color="black",
            markersize=2,
            label=self.baseline_name,
        )
        plt.plot(
            self.days_cut,
            self.running_pt_estimate + 1.65 * self.running_sigmas,
            ".",
            color=sea[2],
            markersize=2,
        )
        plt.plot(
            self.days_cut,
            self.running_pt_estimate - 1.65 * self.running_sigmas,
            ".",
            color=sea[0],
            markersize=2,
        )
        plt.grid(True)
        plt.xlim(self.days_cut[0], self.days_cut[-1])
        if ymin and ymax:
            plt.ylim(ymin, ymax)
        plt.xlabel(self.xaxis, size=self.axes_labelsize)
        plt.ylabel(r"Point estimate $\pm 1.65 \sigma$", size=self.axes_labelsize)
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.annotate(
            f"baseline time: {float(f'{self.percent_obs_segs:.2g}'):g}%",
            xy=(0.5, 0.9),
            xycoords="axes fraction",
            size = self.annotate_fontsize,
            bbox=dict(boxstyle="round", facecolor="white", alpha=1),
        )
        plt.title(f'Running point estimate in {self.time_tag}', fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-running_point_estimate.png", bbox_inches='tight'
        )
        plt.close()

    def plot_running_sigma(self):
        """
        Generates and saves a plot of the running sigma. The plotted values are the ones after the delta sigma cut. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `days_cut`).

        """
        if self.days_cut.size==0:
            return
        fig = plt.figure(figsize=(10, 8))
        plt.plot(
            self.days_cut, self.running_sigmas, '.', markersize=2, color=sea[0], label=self.baseline_name
        )
        plt.grid(True)
        plt.yscale("log")
        plt.xlim(self.days_cut[0], self.days_cut[-1])
        plt.xlabel(self.xaxis, size=self.axes_labelsize)
        plt.ylabel(r"$\sigma$", size=self.axes_labelsize)
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(r'Running $\sigma$ ' + f'in {self.time_tag}', fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-running_sigma.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_IFFT_point_estimate_integrand(self):
        """
        Generates and saves a plot of the IFFT of the point estimate integrand. The IFFT of the point estimate integrand is computed using the method "compute_ifft_integrand". This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_integrand`).
        """
        t_array, omega_array = self.compute_ifft_integrand()
        if len(t_array) != len(omega_array):
            warnings.warn("Times and Omega arrays don't match in the IFFT. No plot could be generated. Investigation is highly recommended.")
            return

        fig = plt.figure(figsize=(10, 8))
        plt.plot(t_array, omega_array, color=sea[0], label=self.baseline_name)
        plt.grid(True)
        plt.xlim(t_array[0], t_array[-1])
        plt.xlabel("Lag (s)", size=self.axes_labelsize)
        plt.ylabel(r"$\Omega$ integrand IFFT", size=self.axes_labelsize)
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(r"$\Omega$ integrand IFFT" + f" in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-IFFT_point_estimate_integrand.png", bbox_inches='tight'
        )
        plt.close()

    def plot_SNR_spectrum(self):
        """
        Generates and saves a plot of the point estimate integrand. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_integrand`).

        """
        if np.isnan(self.point_estimate_spectrum).all() or not np.real(self.point_estimate_spectrum).any():
            return
        fig, axs = plt.subplots(figsize=(10, 8))
        axs.semilogy(
            self.frequencies,
            abs(self.point_estimate_spectrum / self.sigma_spectrum),
            color=sea[0],
        )
        trans = mt.blended_transform_factory(axs.transData, axs.transAxes) 
        axs.vlines(self.frequencies[~self.frequency_mask], ymin=0, ymax=1, linewidth=1, linestyle=':', color='black', alpha=0.5, transform=trans)
        axs.set_xlabel("Frequency (Hz)", size=self.axes_labelsize)
        axs.set_ylabel(r"$|{\rm SNR}(f)|$", size=self.axes_labelsize)
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        axs.set_xscale("log")
        plt.title(f"|SNR| in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-abs_point_estimate_integrand.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_cumulative_SNR_spectrum(self):
        """
        Generates and saves a plot of the cumulative point estimate integrand. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_integrand`).
        """
        pt_est_cumul = self.point_estimate_spectrum.copy()
        pt_est_cumul[~self.frequency_mask] = 0
        cum_pt_estimate = integrate.cumtrapz(
            np.abs(pt_est_cumul/ self.sigma_spectrum), self.frequencies
        )
        cum_pt_estimate = cum_pt_estimate / cum_pt_estimate[-1]
        fig, axs = plt.subplots(figsize=(10, 8))
        axs.plot(self.frequencies[:-1], cum_pt_estimate, color=sea[0])
        axs.set_xlabel("Frequency (Hz)", size=self.axes_labelsize)
        axs.set_ylabel(r"Cumulative $|{\rm SNR}(f)|$", size=self.axes_labelsize)
        axs.set_xscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(f"Cumulative SNR in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-cumulative_SNR_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_real_SNR_spectrum(self):
        """
        Generates and saves a plot of the real part of the SNR spectrum. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_spectrum` and `sigma_spectrum`).
        """
        fig, axs =  plt.subplots(figsize=(10, 8))
        axs.plot(
            self.frequencies,
            np.real(self.point_estimate_spectrum / self.sigma_spectrum),
            color=sea[0],
        )
        trans = mt.blended_transform_factory(axs.transData, axs.transAxes) 
        axs.vlines(self.frequencies[~self.frequency_mask], ymin=0, ymax=1, linewidth=1, linestyle=':', color='black', alpha=0.5, transform=trans)
        axs.set_xlabel("Frequency (Hz)", size=self.axes_labelsize)
        axs.set_ylabel(r"Re$({\rm SNR}(f))$", size=self.axes_labelsize)
        axs.set_xscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(f"Real SNR in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-real_SNR_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_imag_SNR_spectrum(self):
        """
        Generates and saves a plot of the imaginary part of the SNR spectrum. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `point_estimate_spectrum` and `sigma_spectrum`).
        """
        fig, axs = plt.subplots(figsize=(10, 8))
        axs.plot(
            self.frequencies,
            np.imag(self.point_estimate_spectrum / self.sigma_spectrum),
            color=sea[0],
        )
        trans = mt.blended_transform_factory(axs.transData, axs.transAxes) 
        axs.vlines(self.frequencies[~self.frequency_mask], ymin=0, ymax=1, linewidth=1, linestyle=':', color='black', alpha=0.5, transform=trans)
        axs.set_xlabel("Frequency (Hz)", size=self.axes_labelsize)
        axs.set_ylabel(r"Im$({\rm SNR}(f))$", size=self.axes_labelsize)
        axs.set_xscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(f"Imaginary SNR in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-imag_SNR_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_sigma_spectrum(self):
        """
        Generates and saves a plot of the sigma spectrum. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sigma_spectrum`).
        """
        if np.isinf(self.sigma_spectrum).all() or not np.real(self.point_estimate_spectrum).any():
            return

        fig, axs = plt.subplots(figsize=(10, 8))
        axs.plot(self.frequencies, self.sigma_spectrum, color=sea[0])
        trans = mt.blended_transform_factory(axs.transData, axs.transAxes) 
        axs.vlines(self.frequencies[~self.frequency_mask], ymin=0, ymax=1, linewidth=1, linestyle=':', color='black', alpha=0.5, transform=trans)
        axs.set_xlabel("Frequency (Hz)", size=self.axes_labelsize)
        axs.set_ylabel(r"$\sigma(f)$", size=self.axes_labelsize)
        axs.set_xscale("log")
        axs.set_yscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(r"Total $\sigma$ spectrum" + f" in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-sigma_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_coherence_spectrum(self, flow=None, fhigh=None):
        """
        Generates and saves a plot of the coherence spectrum, if present. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `coherence_spectrum`).
        """
        if self.coherence_spectrum is None or self.coherence_spectrum.size==1:
            return

        flow = flow or self.flow
        fhigh = fhigh or self.fhigh

        plt.figure(figsize=(10, 8))
        plt.plot(self.frequencies, self.coherence_spectrum, color=sea[0])
        plt.axhline(y=1./self.n_segs,dashes=(4,3),color='black')
        plt.xlim(flow, fhigh)
        plt.xlabel("Frequency (Hz)", size=self.axes_labelsize)
        plt.ylabel(r"coherence spectrum", size=self.axes_labelsize)
        plt.xscale("log")
        plt.yscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.annotate(
            f"{self.params.channel}",
            xy=(0.01, 0.03),
            xycoords="axes fraction",
            size = self.annotate_fontsize,
            bbox=dict(boxstyle="round", facecolor="white", alpha=1),
        )
        plt.title(r"Coherence ($\Delta f$ = " + f"{float(f'{self.deltaF:.4g}'):g}Hz) in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-coherence_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

        plt.figure(figsize=(10, 8))
        plt.plot(self.frequencies, self.coherence_spectrum, color=sea[0])
        plt.axhline(y=1./self.n_segs,dashes=(4,3),color='black')
        plt.xlim(flow, 200)
        plt.xlabel("Frequency (Hz)", size=self.axes_labelsize)
        plt.ylabel(r"coherence spectrum", size=self.axes_labelsize)
        plt.yscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(r"Coherence ($\Delta f$ = " + f"{float(f'{self.deltaF:.4g}'):g}Hz) in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-coherence_spectrum_zoom.png",
            bbox_inches="tight",
        )
        plt.close()


    def plot_hist_coherence(self,total_bins = None):
        r"""
        Generates and saves a histogram of the coherence distribution. The plot shows the data after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class.
        Furthermore, it also saves a text file which contains the frequencies at which outliers of the coherence distribution were identified, i.e. spectral artefacts.
        """
        if self.coherence_spectrum is None or self.coherence_spectrum.size==1:
            return

        coherence = self.coherence_spectrum
        coherence_clipped = np.ones(len(coherence))
        clip_val = 50* 1/self.n_segs
        for i in range(len(coherence_clipped)):
            if coherence[i] >= clip_val:
                coherence_clipped[i] = clip_val
            else:
                coherence_clipped[i] = coherence[i]
        frequencies = self.frequencies
        if total_bins is None:
            total_bins = 250
        bins =  np.linspace(0, max(coherence), total_bins)
        bins_clipped =  np.linspace(0, max(coherence_clipped), total_bins)
        n_frequencies = len(frequencies)
        delta_coherence = bins[1]-bins[0]
        delta_coherence_clipped = bins_clipped[1]-bins_clipped[0]
        resolution = frequencies[1] - frequencies[0]
        fftlength = int(1.0 / resolution)
        
        coherence_notched = coherence*self.frequency_mask
        coherence_notched_clipped = coherence_clipped*self.frequency_mask

        coherence_highres = np.arange(0,1,1e-6)
        predicted_highres = (self.n_segs-1) * (1- coherence_highres)**(self.n_segs-2)
        threshold = coherence_highres[np.where(predicted_highres <= 1/(n_frequencies*delta_coherence_clipped))[0][0]]

        probability = n_frequencies*delta_coherence_clipped * predicted_highres

        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))
   
        axs.hist(
            coherence,
            bins,
            color=sea[3],
            ec="k",
            lw = 0.1,
            zorder=1,
            density = True,
            label = 'Before notching',
        )
        axs.hist(
            coherence_notched,
            bins,
            color=sea[0],
            ec="k",
            lw = 0.1,
            zorder=2,
            density = True,
            label = 'After notching',
        )
        axs.plot(
            coherence_highres, 
            predicted_highres,
            color=sea[1],
            zorder=3,
            alpha = 0.8,
            label="Predicted",
        )
        axs.axvline(
            np.abs(threshold),
            zorder=4,
            color=sea[8],
            linestyle='dashed',
            label="Threshold",
        )

        axs.set_xlabel(r"Coherence", size=self.axes_labelsize)
        axs.set_ylabel(r"Probability distribution", size=self.axes_labelsize)
        axs.legend(fontsize=self.legend_fontsize)
        axs.set_yscale("log")
        axs.set_xlim(0, max(coherence))
        axs.set_ylim(0.5/(n_frequencies*delta_coherence),10*predicted_highres[0])
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)
        plt.title(r"Coherence hist ($\Delta f$ = " + f"{resolution:.3f}Hz) in" f" {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-histogram_coherence.png", bbox_inches = 'tight'
        )
        plt.close()

        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))      

        axs.hist(
            coherence_clipped,
            bins_clipped,
            color=sea[3],
            ec="k",
            lw = 0.1,
            zorder=1,
            density = True,
            label = 'Before notching',
        )
        axs.hist(
            coherence_notched_clipped,
            bins_clipped,
            color=sea[0],
            ec="k",
            lw = 0.1,
            zorder=2,
            density = True,
            label = 'After notching',
        )
        axs.plot(
            coherence_highres, 
            predicted_highres,
            color=sea[1],
            zorder=3,
            alpha = 0.8,
            label="Predicted",
        )
        axs.axvline(
            np.abs(threshold),
            zorder=4,
            color=sea[8],
            linestyle='dashed',
            label="Threshold",
        )

        axs.set_xlabel(r"Coherence", size=self.axes_labelsize)
        axs.set_ylabel(r"Probability distribution", size=self.axes_labelsize)
        axs.legend(fontsize=self.legend_fontsize)
        axs.set_yscale("log")
        axs.set_xlim(0,max(coherence_clipped))
        axs.set_ylim(0.5/(n_frequencies*delta_coherence_clipped),10*predicted_highres[0])
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)

        plt.title(r"Coherence hist (zoomed) ($\Delta f$ = " + f"{resolution:.3f}Hz) in" f" {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-histogram_coherence_zoom.png", bbox_inches = 'tight'
        )
        plt.close()

        outlier_coherence = [(frequencies[i], coherence[i],probability[np.where(coherence_highres>=coherence[i])[0][0]]) for i in range(len(coherence)) if (coherence[i] > np.abs(threshold) and self.frequency_mask[i] == True)]
        outlier_coherence_notched = [(frequencies[i], coherence[i],probability[np.where(coherence_highres>=coherence[i])[0][0]]) for i in range(len(coherence)) if (coherence[i] > np.abs(threshold) and self.frequency_mask[i] == False)]
        n_outlier = len(outlier_coherence)
        file_name = f"{self.plot_dir / self.baseline_name}-{self.file_tag}-list_coherence_outlier.txt"
        with open(file_name, 'w') as f:
            f.write('Frequencies  \tCoherence \tProbability\n')
            for tup in outlier_coherence:
                f.write(f'{tup[0]}\t{tup[1]}\t{tup[2]}\n')
            f.write('\n The outliers below are already included in the applied version of the notch-list\n')
            for tup in outlier_coherence_notched:
                f.write(f'{tup[0]}\t{tup[1]}\t{tup[2]}\n')
                
    def plot_cumulative_sensitivity(self):
        """
        Generates and saves a plot of the cumulative sensitivity. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sigma_spectrum`).

        """
        if np.isinf(self.sigma_spectrum).all() or not np.real(self.point_estimate_spectrum).any():
            return

        sigma_cumul = self.sigma_spectrum.copy()
        sigma_cumul[~self.frequency_mask] = np.inf
        cumul_sens = integrate.cumtrapz((1 / sigma_cumul ** 2), self.frequencies)
        cumul_sens = cumul_sens / cumul_sens[-1]
        plt.figure(figsize=(10, 8))
        plt.plot(self.frequencies[:-1], cumul_sens, color=sea[0])
        plt.xlabel("Frequency (Hz)", size=self.axes_labelsize)
        plt.ylabel("Cumulative sensitivity", size=self.axes_labelsize)
        plt.xscale("log")
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(r"$1/\sigma^2$ " + f"in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-cumulative_sigma_spectrum.png",
            bbox_inches="tight",
        )
        plt.close()

    def plot_omega_sigma_in_time(self):
        r"""
        Generates and saves a panel plot with a scatter plot of :math:`\sigma` vs :math:`\Delta{\rm SNR}_i`, as well as the evolution of :math:`\Omega`, :math:`\sigma`, and :math:`(\Omega-\langle\Omega\rangle)/\sigma` as a function of the days since the start of the run. All plots show the data before and after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_sigmas_all`).
        """
        fig, axs = plt.subplots(nrows=3, ncols=1, figsize=(10, 15), constrained_layout=True)
        fig.suptitle(r"$\Omega$, $\sigma$, and" + f" SNR variations in {self.time_tag} with/out " + r"$\Delta\sigma$ cut", fontsize=self.title_fontsize)

        axs[0].plot(self.days_all, self.sliding_omega_all, color=sea[3], linewidth=1, alpha=0.5, label="All data")
        axs[0].plot(
            self.days_cut,
            self.sliding_omega_cut,
            color=sea[0], linewidth=1, alpha=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
        )
        axs[0].plot(self.days_all, self.sliding_omega_all, '.', color=sea[3])
        axs[0].plot(
            self.days_cut,
            self.sliding_omega_cut, '.',
            color=sea[0],
        )
        axs[0].set_xlabel(self.xaxis, size=self.axes_labelsize)
        axs[0].set_ylabel(r"$\Omega$", size=self.axes_labelsize)
        axs[0].legend(loc="upper left", fontsize=self.legend_fontsize)
        axs[0].set_xlim(0, self.days_all[-1])
        axs[0].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[0].tick_params(axis="y", labelsize=self.legend_fontsize)
        axs[0].yaxis.offsetText.set_fontsize(self.legend_fontsize)

        axs[1].plot(self.days_all, self.sliding_sigmas_all, color=sea[3], linewidth=1, alpha=0.5, label="All data")
        axs[1].plot(
            self.days_cut,
            self.sliding_sigma_cut,
            color=sea[0],
            linewidth=1,
            alpha=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
        )
        axs[1].plot(self.days_all, self.sliding_sigmas_all,'.', color=sea[3])
        axs[1].plot(
            self.days_cut,
            self.sliding_sigma_cut,'.',
            color=sea[0]
        )
        axs[1].set_xlabel(self.xaxis, size=self.axes_labelsize)
        axs[1].set_ylabel(r"$\sigma$", size=self.axes_labelsize)
        axs[1].legend(loc="upper left", fontsize=self.legend_fontsize)
        axs[1].set_xlim(0, self.days_all[-1])
        axs[1].set_yscale('log')
        axs[1].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[1].tick_params(axis="y", labelsize=self.legend_fontsize)
        axs[1].yaxis.offsetText.set_fontsize(self.legend_fontsize)

        axs[2].plot(self.days_all, self.sliding_deviate_all, color=sea[3], linewidth=1, alpha=0.5, label="All data")
        axs[2].plot(
            self.days_cut,
            self.sliding_deviate_cut,
            color=sea[0], linewidth=1, alpha=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
        )
        axs[2].plot(self.days_all, self.sliding_deviate_all, '.', color=sea[3])
        axs[2].plot(
            self.days_cut,
            self.sliding_deviate_cut, '.',
            color=sea[0],
        )
        axs[2].set_xlabel(self.xaxis, size=self.axes_labelsize)
        axs[2].set_ylabel(r"$\Delta{\rm SNR}_i$", size=self.axes_labelsize)
        axs[2].legend(loc="upper left", fontsize=self.legend_fontsize)
        axs[2].set_xlim(0, self.days_all[-1])
        axs[2].set_yscale("symlog")
        axs[2].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[2].tick_params(axis="y", labelsize=self.legend_fontsize)

        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-omega_sigma_time.png", bbox_inches='tight'
        )
        plt.close()

    def plot_hist_sigma_dsc(self):
        r"""
        Generates and saves a panel plot with a histogram of :math:`|\Delta\sigma|/\sigma`, as well as a histogram of :math:`\sigma`. Both plots show the data before and after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `delta_sigmas_all`).

        """
        fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(10, 14), constrained_layout=True)
        fig.suptitle(r"$\Delta\sigma$ and $\sigma$ distributions in" f" {self.time_tag} with/out " + r"$\Delta\sigma$ cut", fontsize=self.title_fontsize)

        axs[0].hist(
            self.delta_sigmas_all,
            bins=80,
            color=sea[3],
            ec="k",
            lw=0.5,
            label="All data",
            range=(0.0001, 1),
        )
        axs[0].hist(
            self.delta_sigmas_cut,
            bins=80,
            color=sea[0],
            alpha = 0.6,
            ec="k",
            lw=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
            range=(0.0001, 1),
        )
        axs[0].set_xlabel(r"$|\Delta\sigma|/\sigma$", size=self.axes_labelsize)
        axs[0].set_ylabel(r"count", size=self.axes_labelsize)
        axs[0].legend(fontsize=self.legend_fontsize)
        axs[0].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[0].tick_params(axis="y", labelsize=self.legend_fontsize)
        axs[0].yaxis.offsetText.set_fontsize(self.legend_fontsize)

        if self.sliding_sigma_cut.size==0:
            minx1 = min(self.sliding_sigmas_all)
            maxx1 = max(self.sliding_sigmas_all)
        else:
            minx1 = min(self.sliding_sigma_cut)
            maxx1 = max(self.sliding_sigma_cut)
        nx = 50

        axs[1].hist(
            self.sliding_sigmas_all,
            bins=nx,
            color=sea[3],
            ec="k",
            lw=0.5,
            label="All data",
            range=(minx1, maxx1),
        )
        axs[1].hist(
            self.sliding_sigma_cut,
            bins=nx,
            color=sea[0],
            alpha = 0.6,
            ec="k",
            lw=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
            range=(minx1, maxx1),
        )
        axs[1].set_xlabel(r"$\sigma$", size=self.axes_labelsize)
        axs[1].set_ylabel(r"count", size=self.axes_labelsize)
        axs[1].legend(fontsize=self.legend_fontsize)
        axs[1].set_yscale("log")
        axs[1].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[1].tick_params(axis="y", labelsize=self.legend_fontsize)
        axs[1].xaxis.offsetText.set_fontsize(self.legend_fontsize)

        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-histogram_sigma_dsc.png", bbox_inches='tight'
        )
        plt.close()

    def plot_scatter_sigma_dsc(self):
        """
        Generates and saves a scatter plot of :math:`|\Delta\sigma]/\sigma` vs :math:`\sigma`. The plot shows the data before and after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `delta_sigmas_all`).
        """
        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))

        axs.scatter(
            self.delta_sigmas_all,
            self.naive_sigmas_all,
            marker=".",
            color=sea[3],
            label="All data",
            s=5,
        )
        axs.scatter(
            self.delta_sigmas_cut,
            self.naive_sigma_cut,
            marker=".",
            color=sea[0],
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",

            s=5,
        )
        axs.set_xlabel(r"$|\Delta\sigma|/\sigma$", size=self.axes_labelsize)
        axs.set_ylabel(r"$\sigma$", size=self.axes_labelsize)
        axs.set_yscale("log")
        axs.set_xscale("log")
        axs.legend(fontsize=self.legend_fontsize)
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)

        axs.annotate(
            r"Data cut by $\Delta\sigma$ cut"+f": {float(f'{self.dsc_percent:.2g}'):g}%",
            xy=(0.05, 0.8),
            xycoords="axes fraction",
            size = self.annotate_fontsize,
            bbox=dict(boxstyle="round", facecolor="white", alpha=1),
        )
        plt.title(r"$\Delta\sigma$ distribution in" f" {self.time_tag} with/out " + r"$\Delta\sigma$ cut", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-scatter_sigma_dsc.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_scatter_omega_sigma_dsc(self):
        r"""
        Generates and saves a panel plot with scatter plots of :math:`|\Delta\sigma|/\sigma` vs :math:`\Delta{\rm SNR}_i`, as well as :math:`\sigma` vs :math:`(\Omega-\langle\Omega\rangle)/\sigma`. All plots show the data before and after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `delta_sigmas_all`).
        """
        fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(10, 13), constrained_layout=True)
        fig.suptitle(r"$\Delta$SNR spread" + f" in {self.time_tag} with/out " + r"$\Delta\sigma$ cut", fontsize=self.title_fontsize)

        if self.delta_sigmas_cut.size==0:
            maxx0 = max(self.delta_sigmas_all)
            maxx0 += maxx0 / 10.0
            minx0 = min(self.delta_sigmas_all)
            minx0 -= minx0 / 10.0
            maxy0 = np.nanmax(self.sliding_deviate_all)
            maxy0 += maxy0 / 10.0
            miny0 = np.nanmin(self.sliding_deviate_all)
            miny0 -= miny0 / 10.0

            maxx1 = max(self.sliding_sigmas_all)
            maxx1 += maxx1 / 10.0
            minx1 = min(self.sliding_sigmas_all)
            minx1 -= minx1 / 10.0
            maxy1 = max(self.sliding_deviate_all)
            maxy1 += maxy1 / 10.0
            miny1 = min(self.sliding_deviate_all)
            miny1 -= miny1 / 10.0

        else:
            maxx0 = max(self.delta_sigmas_cut)
            maxx0 += maxx0 / 10.0
            minx0 = min(self.delta_sigmas_cut)
            minx0 -= minx0 / 10.0
            maxy0 = np.nanmax(self.sliding_deviate_cut)
            maxy0 += maxy0 / 10.0
            miny0 = np.nanmin(self.sliding_deviate_cut)
            miny0 -= miny0 / 10.0

            maxx1 = max(self.sliding_sigma_cut)
            maxx1 += maxx1 / 10.0
            minx1 = min(self.sliding_sigma_cut)
            minx1 -= minx1 / 10.0
            maxy1 = max(self.sliding_deviate_cut)
            maxy1 += maxy1 / 10.0
            miny1 = min(self.sliding_deviate_cut)
            miny1 -= miny1 / 10.0

        axs[0].scatter(
            self.delta_sigmas_all,
            self.sliding_deviate_all,
            marker=".",
            color=sea[3],
            label="All data",
            s=3,
        )
        axs[0].scatter(
            self.delta_sigmas_cut,
            self.sliding_deviate_cut,
            marker=".",
            color=sea[0],
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
            s=3,
        )
        axs[0].set_xlabel(r"$|\Delta\sigma|/\sigma$", size=self.axes_labelsize)
        axs[0].set_ylabel(r"$\Delta{\rm SNR}_i$", size=self.axes_labelsize)
        axs[0].set_xlim(minx0, maxx0)
        axs[0].set_ylim(miny0, maxy0)
        axs[0].legend(fontsize=self.legend_fontsize)
        axs[0].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[0].tick_params(axis="y", labelsize=self.legend_fontsize)


        axs[1].scatter(
            self.sliding_sigmas_all,
            self.sliding_deviate_all,
            marker=".",
            color=sea[3],
            label="All data",
            s=3,
        )
        axs[1].scatter(
            self.sliding_sigma_cut,
            self.sliding_deviate_cut,
            marker=".",
            color=sea[0],
            label=r"Data after $|\Delta\sigma/\sigma|$ outlier cut",
            s=3,
        )
        axs[1].set_xlabel(r"$\sigma$", size=self.axes_labelsize)
        axs[1].set_ylabel(r"$\Delta{\rm SNR}_i$", size=self.axes_labelsize)
        axs[1].legend(fontsize=self.legend_fontsize)
        axs[1].set_xlim(minx1, maxx1)
        axs[1].set_ylim(miny1, maxy1)
        axs[1].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[1].tick_params(axis="y", labelsize=self.legend_fontsize)
        axs[1].xaxis.offsetText.set_fontsize(self.legend_fontsize)

        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-scatter_omega_sigma_dsc.png", bbox_inches = 'tight')
        plt.close()

    def plot_hist_omega_pre_post_dsc(self):
        r"""
        Generates and saves a histogram of the :math:`\Delta{\rm SNR}_i` distribution. The plot shows the data before and after the delta-sigma cut (bad GPS times) was applied. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_deviate_all`).
        """
        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))

        # nan-safing the histograms for good measure...
        bins=np.histogram(np.hstack((self.sliding_deviate_all[~np.isnan(self.sliding_deviate_all)], self.sliding_deviate_cut[~np.isnan(self.sliding_deviate_cut)])), bins=202)[1]
        
        axs.hist(
            self.sliding_deviate_all,
            bins,
            color=sea[3],
            ec="k",
            lw=0.5,
            label="All data",
        )
        axs.hist(
            self.sliding_deviate_cut,
            bins,
            color=sea[0],
            alpha = 0.6,
            ec="k",
            lw=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
        )
        
        axs.set_xlabel(r"$\Delta{\rm SNR}_i$", size=self.axes_labelsize)
        axs.set_ylabel(r"count", size=self.axes_labelsize)
        axs.legend(fontsize=self.legend_fontsize)
        axs.set_yscale("log")
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)

        plt.title(r"$\Delta$SNR distribution in" f" {self.time_tag} with/out " + r"$\Delta\sigma$ cut", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-histogram_omega_dsc.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_KS_test(self, bias_factor=None):
        """
        Generates and saves a panel plot with results of the Kolmogorov-Smirnov test for Gaussianity. The cumulative distribution of the data (after the delta-sigma (bad GPS times) cut) is compared to the one of Gaussian data, where the bias factor for the sigmas is taken into account. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_deviate_cut`).

        Parameters
        ==========
        bias_factor: float
            Bias factor to consider in the KS calculation.

        """
        if self.delta_sigmas_cut.size==0:
            return

        if bias_factor is None:
            bias_factor = calc_bias(self.segment_duration, self.deltaF, self.deltaT)
        dof_scale_factor = 1.0 / (1.0 + 3.0 / 35.0)
        lx = len(self.sliding_deviate_cut)

        sorted_deviates = np.sort(self.sliding_deviate_KS / bias_factor)

        nanmask = ~np.isnan(sorted_deviates)
        sorted_deviates_nansafe = sorted_deviates[nanmask]
        count, bins_count = np.histogram(sorted_deviates_nansafe, bins=500)
        pdf = count / sum(count)
        cdf = np.cumsum(pdf)

        normal = stats.norm(0, 1)
        normal_cdf = normal.cdf(bins_count[1:])

        dks_x = max(abs(cdf - normal_cdf))
        lx_eff = lx * dof_scale_factor

        lam = (np.sqrt(lx_eff) + 0.12 + 0.11 / np.sqrt(lx_eff)) * dks_x
        pval_KS = StatKS(lam)

        fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(10, 8), constrained_layout=True)
        fig.suptitle(f"Kolmogorov-Smirnov test in {self.time_tag}", fontsize=self.title_fontsize)

        axs[0].plot(bins_count[1:], cdf, "k", label="Data")
        axs[0].plot(
            bins_count[1:],
            normal_cdf,
            color=sea[3],
            label=r"Erf with $\sigma$=" + str(round(bias_factor, 2)),
        )
        axs[0].text(
            bins_count[1],
            0.8,
            "KS-statistic: "
            + str(round(dks_x, 3))
            + "\n"
            + "p-value: "
            + str(round(pval_KS, 3)),
        )
        axs[0].legend(loc="lower right", fontsize=self.legend_fontsize)
        axs[0].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[0].tick_params(axis="y", labelsize=self.legend_fontsize)

        axs[1].plot(
            bins_count[1:],
            cdf - normal_cdf,
        )
        axs[1].annotate(
            "Maximum absolute difference: " + str(round(dks_x, 3)),
            xy=(0.025, 0.9),
            xycoords="axes fraction",
            size = self.annotate_fontsize
        )
        axs[1].tick_params(axis="x", labelsize=self.legend_fontsize)
        axs[1].tick_params(axis="y", labelsize=self.legend_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-KS_test.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_hist_sigma_squared(self, max_val = 5, total_bins=100, label_number = 6):
        """
        Generates and saves a histogram of :math:`\sigma^2/\langle\sigma^2\rangle`. The plot shows data after the delta-sigma (bad GPS times) cut. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. self.sliding_sigma_cut).

        Parameters
        ==========

        Returns
        =======

        """
        if self.delta_sigmas_cut.size==0:
            return
        
        values = 1 / np.nanmean(self.sliding_sigma_cut ** 2) * self.sliding_sigma_cut ** 2
        
        values_clipped = np.ones(len(values))
        
        for i in range(len(values_clipped)):
            if values[i] >= max_val:
                values_clipped[i] = max_val
            else:
                values_clipped[i] = values[i]
                
        bins = np.linspace(0, max(values_clipped), total_bins)
        
        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))
        _, bins, patches = plt.hist(
            values_clipped,
            bins=bins,
            color=sea[0],
            ec="k",
            lw=0.5,
            label=r"Data after $|\Delta\sigma|/\sigma$ outlier cut",
        )
        
        axs.set_xlabel(r"$\sigma^2/\langle\sigma^2\rangle$", size=self.axes_labelsize)
        axs.set_ylabel(r"count", size=self.axes_labelsize)
        axs.set_yscale("log")
        axs.set_xlim(0, max_val)
        axs.legend(fontsize=self.legend_fontsize)
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)
        
        xticks_tmp = np.linspace(0, max_val, label_number)
        labels = [str(i) for i in xticks_tmp]
        labels[-1]+="+"
        plt.xticks(xticks_tmp, labels)
        
        plt.title(f"Relative variance in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-histogram_sigma_squared.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_omega_time_fit(self):
        """
        Generates and saves a plot of :math:`\Omega` as a function of time and fits the data to perform a linear trend analysis. The plot shows data after the delta-sigma (bad GPS times) cut. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. self.sliding_omega_cut).

        Parameters
        ==========

        Returns
        =======

        """
        if self.days_cut.size==0:
            return
        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))

        t_obs = self.days_cut[-1]
        scale = np.sqrt(np.var(self.sliding_omega_cut))

        def func(x, a, b):
            return a * (x - t_obs / 2) * t_obs + b

        nanmask = ~np.isnan(self.sliding_omega_cut)
        sliding_omega_cut_nansafe = self.sliding_omega_cut[nanmask]
        sliding_sigma_cut_nansafe = self.sliding_sigma_cut[nanmask]

        popt, pcov = curve_fit(
            func,
            self.days_cut[nanmask],
            sliding_omega_cut_nansafe,
            sigma=sliding_sigma_cut_nansafe,
        )
        c1, c2 = popt[0], popt[1]
        axs.plot(self.days_cut, func(self.days_cut, c1, c2), color=sea[3])
        axs.plot(self.days_cut, self.sliding_omega_cut, '.', color=sea[0], markersize=1)
        axs.plot(self.days_cut, 3 * self.sliding_sigma_cut, color=sea[0], linewidth=1.5)
        axs.plot(self.days_cut, -3 * self.sliding_sigma_cut, color=sea[0], linewidth=1.5)
        axs.set_xlabel(self.xaxis, size=self.axes_labelsize)
        axs.set_ylabel(r"$\Omega_i$", size=self.axes_labelsize)
        axs.set_xlim(self.days_cut[0], self.days_cut[-1])
        axs.annotate(
            r"Linear trend analysis: $\Omega(t) = C_1 (t-T_{\rm obs}/2) T_{\rm obs} + C_2 C_1 = $"
            + str(f"{c1:.3e}")
            + "\nC$_2$ = "
            + str(f"{c2:.3e}"),
            xy=(0.05, 0.05),
            xycoords="axes fraction",
            size = self.annotate_fontsize,
            bbox=dict(boxstyle="round", facecolor="white", alpha=1),
        )
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)
        axs.xaxis.offsetText.set_fontsize(self.legend_fontsize)
        plt.title(r"Time evolution of $\Omega$ " + f"in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-omega_time_fit.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_sigma_time_fit(self):
        """
        Generates and saves a plot of :math:`\sigma` as a function of time and fits the data to perform a linear trend analysis. The plot shows data after the delta-sigma (bad GPS times) cut. This function does not require any input parameters, as it accesses the data through the attributes of the class (e.g. `sliding_sigma_cut`).
        """
        if self.days_cut.size==0:
            return

        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))

        t_obs = self.days_cut[-1]
        scale = np.sqrt(np.var(self.sliding_sigma_cut))

        def func(x, a, b):
            return a * (x - t_obs / 2) * t_obs + b

        nanmask = ~np.isnan(self.sliding_sigma_cut)
        sliding_sigma_cut_nansafe = self.sliding_sigma_cut[nanmask]
        mean_sigma = np.nanmean(self.sliding_sigma_cut)

        popt, pcov = curve_fit(
            func,
            self.days_cut[nanmask],
            sliding_sigma_cut_nansafe,
        )
        c1, c2 = popt[0], popt[1]
        axs.plot(self.days_cut, func(self.days_cut, c1, c2), color=sea[3])
        axs.plot(self.days_cut, self.sliding_sigma_cut, ".", color=sea[0], markersize=1)
        axs.set_xlabel(self.xaxis, size=self.axes_labelsize)
        axs.set_ylabel(r"$\sigma_i$", size=self.axes_labelsize)
        axs.set_xlim(self.days_cut[0], self.days_cut[-1])
        axs.set_ylim(mean_sigma - 1.2 * mean_sigma, mean_sigma + 2.2 * mean_sigma)
        axs.annotate(
            r"Linear trend analysis: $\sigma(t) = C_1 (t-T_{\rm obs}/2) T_{\rm obs} + C_2 C_1 = $"
            + str(f"{c1:.3e}")
            + "\nC$_2$ = "
            + str(f"{c2:.3e}"),
            xy=(0.05, 0.05),
            xycoords="axes fraction",
            size = self.annotate_fontsize,
            bbox=dict(boxstyle="round", facecolor="white", alpha=1),
        )
        axs.tick_params(axis="x", labelsize=self.legend_fontsize)
        axs.tick_params(axis="y", labelsize=self.legend_fontsize)
        plt.title(r"Time evolution of $\sigma$ " + f"in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-sigma_time_fit.png", bbox_inches = 'tight'
        )
        plt.close()

    def plot_gates_in_time(self):
        if self.gates_ifo1 is None and self.gates_ifo2 is None:
            self.gates_ifo1_statement=None
            self.gates_ifo2_statement=None
            return
        
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 8))
        if self.gates_ifo1 is None:
            self.gates_ifo1_statement=None
        else:
            self.total_gated_time_ifo1 = np.sum(self.gates_ifo1[:,1]-self.gates_ifo1[:,0])
            self.total_gated_percent_ifo1 = self.total_gated_time_ifo1/(int(self.sliding_times_all[-1])- int(self.sliding_times_all[0]))*100
            gate_times_in_days_ifo1 = (np.array(self.gates_ifo1[:,0]) - self.sliding_times_all[0]) / 86400.0
            self.gates_ifo1_statement= f"Data gated out: {self.total_gated_time_ifo1} s\n" f"Percentage: {float(f'{self.total_gated_percent_ifo1:.2g}'):g}%"
            gatefig1 = ax.plot(gate_times_in_days_ifo1, self.gates_ifo1[:,1]-self.gates_ifo1[:,0], 's', color=sea[3], label="IFO1:\n" f"{self.gates_ifo1_statement}")
            first_legend = ax.legend(handles=gatefig1, loc=(0.05,0.75), fontsize = self.axes_labelsize)
            ax.add_artist(first_legend)
        if self.gates_ifo2 is None:
            self.gates_ifo2_statement=None
        else:
            self.total_gated_time_ifo2 = np.sum(self.gates_ifo2[:,1]-self.gates_ifo2[:,0])
            self.total_gated_percent_ifo2 = self.total_gated_time_ifo2/(int(self.sliding_times_all[-1])- int(self.sliding_times_all[0]))*100
            gate_times_in_days_ifo2 = (np.array(self.gates_ifo2[:,0]) - self.sliding_times_all[0]) / 86400.0
            self.gates_ifo2_statement= f"Data gated out: {self.total_gated_time_ifo2} s\n" f"Percentage: {float(f'{self.total_gated_percent_ifo2:.2g}'):g}%"
            gatefig2 = ax.plot(gate_times_in_days_ifo2, self.gates_ifo2[:,1]-self.gates_ifo2[:,0], 's', color=sea[0], label="IFO2:\n" f"{self.gates_ifo2_statement}")
            ax.legend(handles=gatefig2, loc=(0.05, 0.1), fontsize = self.axes_labelsize)
        ax.set_xlabel(self.xaxis, size=self.axes_labelsize)
        ax.set_ylabel("Gate length (s)", size=self.axes_labelsize)
        plt.xticks(fontsize=self.legend_fontsize)
        plt.yticks(fontsize=self.legend_fontsize)
        plt.title(f"Gates applied to {self.baseline_name} in {self.time_tag}", fontsize=self.title_fontsize)
        plt.savefig(
            f"{self.plot_dir / self.baseline_name}-{self.file_tag}-gates_time.png",
            bbox_inches="tight",
        )
        plt.close()
            


    def save_all_statements(self):
        """
        Saves all useful statements gathered throughout the checks to a json file.
        """
        statements = {}
        statements['dsc'] = self.dsc_statement
        statements['gates_ifo1'] = self.gates_ifo1_statement
        statements['gates_ifo2'] = self.gates_ifo2_statement
        statements['n_segs'] = self.n_segs_statement
        with open("stats_statements.json", "w") as outfile:
                json.dump(statements, outfile)

    def generate_all_plots(self):
        """
        Generates and saves all the statistical analysis plots.
        """
        self.plot_running_point_estimate()
        self.plot_running_sigma()
        self.plot_IFFT_point_estimate_integrand()
        self.plot_SNR_spectrum()
        self.plot_cumulative_SNR_spectrum()
        self.plot_real_SNR_spectrum()
        self.plot_imag_SNR_spectrum()
        self.plot_sigma_spectrum()
        self.plot_cumulative_sensitivity()
        self.plot_omega_sigma_in_time()
        self.plot_hist_sigma_dsc()
        self.plot_scatter_sigma_dsc()
        self.plot_scatter_omega_sigma_dsc()
        self.plot_hist_omega_pre_post_dsc()
        self.plot_KS_test()
        self.plot_hist_sigma_squared()
        self.plot_omega_time_fit()
        self.plot_sigma_time_fit()
        self.plot_gates_in_time()
        if self.coherence_spectrum is not None:
            self.plot_coherence_spectrum()
            self.plot_hist_coherence()
        self.save_all_statements()


def sortingFunction(item):
    return float(item[5:].partition("-")[0])


def run_statistical_checks_from_file(
    combine_file_path, dsc_file_path, plot_dir, param_file, legend_fontsize=16, coherence_file_path = None, file_tag = None, convention='pygwb'
):
    """
    Assumes files are in npz for now. Will generalize later.
    """
    params = Parameters()
    params.update_from_file(param_file)

    spectra_file = np.load(combine_file_path)
    dsc_file = np.load(dsc_file_path)

    badGPStimes = dsc_file["badGPStimes"]
    delta_sigmas = dsc_file["delta_sigmas"]
    sliding_times = dsc_file["times"]
    naive_sigma_all = dsc_file["naive_sigmas"]
    gates_ifo1 = dsc_file["gates_ifo1"]
    gates_ifo2 = dsc_file["gates_ifo2"]
    if gates_ifo1.size==0:
        gates_ifo1=None
    if gates_ifo2.size==0:
        gates_ifo2=None

    sliding_omega_all, sliding_sigmas_all = (
        spectra_file["point_estimates_seg_UW"],
        spectra_file["sigmas_seg_UW"],
    )

    frequencies = np.arange(
        0,
        params.new_sample_rate / 2.0 + params.frequency_resolution,
        params.frequency_resolution,
    )
    frequency_cut = (params.fhigh >= frequencies) & (frequencies >= params.flow)
    try:
        frequency_mask = spectra_file["frequency_mask"]
    except KeyError:
        try:
            notch_list_path = params.notch_list_path
            logger.debug("loading notches from " + notch_list_path)
            notch_list = StochNotchList.load_from_file(notch_list_path)
            frequency_mask = notch_list.get_notch_mask(frequencies[frequency_cut])

        except:
            frequency_mask = None

    spectrum_file = np.load(combine_file_path, mmap_mode="r")

    point_estimate_spectrum = spectrum_file["point_estimate_spectrum"]
    sigma_spectrum = spectrum_file["sigma_spectrum"]

    baseline_name = params.interferometer_list[0] + params.interferometer_list[1]

    # select alpha for statistical checks
    delta_sigmas_sel = delta_sigmas.T[1]
    naive_sigmas_sel = naive_sigma_all.T[1]

    if coherence_file_path is not None:
        coh_data = np.load(coherence_file_path, allow_pickle=True)
        coherence_spectrum = coh_data['coherence']
        coherence_n_segs = coh_data['n_segs_coh']
    else:
        coherence_spectrum = None
        coherence_n_segs = None

    return StatisticalChecks(
        sliding_times,
        sliding_omega_all,
        sliding_sigmas_all,
        naive_sigmas_sel,
        coherence_spectrum,
        coherence_n_segs,
        point_estimate_spectrum,
        sigma_spectrum,
        frequencies[frequency_cut],
        badGPStimes,
        delta_sigmas_sel,
        plot_dir,
        baseline_name,
        param_file,
        frequency_mask,
        gates_ifo1,
        gates_ifo2,
        file_tag=file_tag,
        legend_fontsize=legend_fontsize,
        convention=convention
    )




from abc import abstractmethod

import bilby
import numpy as np
from scipy.special import erf

from .baseline import Baseline


class GWBModel(bilby.Likelihood):
    r"""
    Generic model, contains the definition of likelihood:

    .. math::

        p(\hat{C}^{IJ}(f_k) | \mathbf{\Theta}) \propto\exp\left[  -\frac{1}{2} \sum_{IJ}^N \sum_k \left(\frac{\hat{C}^{IJ}(f_k) - \Omega_{\rm M}(f_k|\mathbf{\Theta})}{\sigma^2_{IJ}(f_k)}\right)^2  \right],

    where :math:`\Omega_{\rm M}(f_k|\mathbf{\Theta})` is the model being fit to data, and :math:`\mathbf{\Theta}` are the model's parameters.

    The noise likelihood is given by setting :math:`\Omega_{\rm M}(f_k|\mathbf{\Theta})=0`.

    """

    def __init__(self, baselines=None, model_name=None, polarizations=None):
        super(GWBModel, self).__init__()
        # list of baselines
        if baselines is None:
            raise ValueError("list of baselines must be supplied")
        if model_name is None:
            raise ValueError("model_name must be supplied")
        # if single baseline supplied, that's fine
        # just make it a list
        if isinstance(baselines, Baseline):
            baselines = [baselines]
        self.baselines = baselines
        self.orfs = []
        self.polarizations = polarizations
        # if polarizations is not supplied
        if polarizations is None:
            self.polarizations = ["tensor" for ii in range(len(baselines))]
        for bline in self.baselines:
            if self.polarizations[0].lower() == "tensor":
                self.orfs.append(bline.tensor_overlap_reduction_function)
            elif self.polarizations[0].lower() == "vector":
                self.orfs.append(bline.vector_overlap_reduction_function)
            elif self.polarizations[0].lower() == "scalar":
                self.orfs.append(bline.scalar_overlap_reduction_function)
            else:
                raise ValueError(
                    f"unexpected type for polarizations {type(polarizations)}"
                )
        bilby.Likelihood.__init__(self, parameters=self.parameters)

    @abstractmethod
    def parameters(self):
        """Parameters. Should return a dictionary"""
        pass

    @abstractmethod
    def model_function(self):
        """Function for evaluating model"""
        pass

    def log_likelihood_IJ(self, baseline, noise=False):
        """Function for evaluating log likelihood of IJ baseline pair"""
        if noise:
            Y_model_f = 0
        else:
            Y_model_f = self.model_function(baseline)

        # simple likelihood without calibration uncertainty
        if baseline.calibration_epsilon == 0:
            logL_IJ = -0.5 * (
                np.sum(
                    (baseline.point_estimate_spectrum - Y_model_f) ** 2
                    / baseline.sigma_spectrum ** 2
                )
                + np.sum(np.log(2 * np.pi * baseline.sigma_spectrum ** 2))
            )

        # likelihood with calibration uncertainty marginalizatione done analytically
        # see https://stochastic-alog.ligo.org/aLOG//index.php?callRep=339711
        # note \cal{N} = \Prod_j sqrt(2*pi*sigma_j^2)
        else:
            A = baseline.calibration_epsilon ** (-2) + np.sum(
                Y_model_f ** 2 / baseline.sigma_spectrum ** 2
            )
            B = baseline.calibration_epsilon ** (-2) + np.sum(
                Y_model_f
                * baseline.point_estimate_spectrum
                / baseline.sigma_spectrum ** 2
            )
            C = baseline.calibration_epsilon ** (-2) + np.sum(
                baseline.point_estimate_spectrum ** 2 / baseline.sigma_spectrum ** 2
            )
            log_norm = -0.5 * np.sum(np.log(2 * np.pi * baseline.sigma_spectrum ** 2))

            logL_IJ = (
                log_norm
                - 0.5 * np.log(A * baseline.calibration_epsilon ** 2)
                + np.log(1 + erf(B / np.sqrt(2 * A)))
                - np.log(1 + erf(1 / np.sqrt(2 * baseline.calibration_epsilon ** 2)))
                - 0.5 * (C - B ** 2 / A)
            )

        return logL_IJ

    def log_likelihood(self):
        """Function for evaluating log likelihood of detector network"""
        ll = 0
        for baseline in self.baselines:
            ll = ll + self.log_likelihood_IJ(baseline, noise=False)
        return ll

    def noise_log_likelihood(self):
        """Function for evaluating noise log likelihood of detector network"""
        ll = 0
        for baseline in self.baselines:
            ll = ll + self.log_likelihood_IJ(baseline, noise=True)
        return ll


class PowerLawModel(GWBModel):
    r"""
    Power law model is defined as:

    .. math::
        \Omega(f) = \Omega_{\text{ref}} \left(\frac{f}{f_{\text{ref}}}\right)^{\alpha}
    Parameters:
    -----------
    fref : float
        reference frequency for defining the model (:math:`f_{\text{ref}}`)
    omega_ref : float
        amplitude of signal at fref (:math:`\Omega_{\text{ref}}`)
    alpha : float
        spectral index of the power law (:math:`\alpha`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        try:
            fref = kwargs.pop("fref")
        except KeyError:
            raise KeyError("fref must be supplied")
        super(PowerLawModel, self).__init__(**kwargs)
        self.fref = fref

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {"omega_ref": None, "alpha": None}
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        return (
            self.parameters["omega_ref"]
            * (bline.frequencies / self.fref) ** self.parameters["alpha"]
        )


class BrokenPowerLawModel(GWBModel):
    r"""
    Broken Power law model is defined as: 
    
    .. math:: 
        \Omega(f) = \begin{cases}
            \Omega_{\text{ref}} \left( \frac{f}{f_{\text{ref}}} \right) ^ {\alpha_1}, f \leqslant f_{\text{ref}} \\
            \Omega_{\text{ref}} \left( \frac{f}{f_{\text{ref}}} \right) ^ {\alpha_2}, f > f_{\text{ref}}
        \end{cases}
    Parameters:
    -----------
    omega_ref : float
        amplitude of signal at fref (:math:`\Omega_{\text{ref}}`)
    alpha_1 : float
        spectral index of the broken power law (:math:`\alpha_1`)
    alpha_2 : float
        spectral index of the broken power law (:math:`\alpha_2`)
    fbreak : float
        break frequency for the broken power law (:math:`f_{\text{ref}}`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        super(BrokenPowerLawModel, self).__init__(**kwargs)

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {"omega_ref": None, "fbreak": None, "alpha_1": None, "alpha_2": None}
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        frequencies = bline.frequencies
        return self.parameters["omega_ref"] * (
            np.piecewise(
                frequencies,
                [
                    frequencies <= self.parameters["fbreak"],
                    frequencies > self.parameters["fbreak"],
                ],
                [
                    lambda frequencies: (frequencies / self.parameters["fbreak"])
                    ** (self.parameters["alpha_1"]),
                    lambda frequencies: (frequencies / self.parameters["fbreak"])
                    ** (self.parameters["alpha_2"]),
                ],
            )
        )


class TripleBrokenPowerLawModel(GWBModel):
    r"""
    The triple broken power law is defined as: 
    
    .. math:: 
        \Omega(f) = \begin{cases}
            \Omega_{\text{ref}} \left( \frac{f}{f_1} \right) ^ {\alpha_1}, f \leqslant f_1 \\
            \Omega_{\text{ref}} \left( \frac{f}{f_1} \right) ^ {\alpha_2}, f_1 < f \leqslant f_2  \\
            \Omega_{\text{ref}} \left( \frac{f_2}{f_1} \right) ^{\alpha_2} \left( \frac{f}{f_2} \right)^{\alpha_3}, f > f_2
        \end{cases}
    Parameters:
    -----------
    omega_ref : float
        amplitude of signal at fref (:math:`\Omega_{\text{ref}}`) 
    alpha_1 : float
        spectral index of the broken power law (:math:`\alpha_1`)
    alpha_2 : float
        spectral index of the broken power law (:math:`\alpha_2`)
    alpha_3 : float
        spectral index of the broken power law (:math:`\alpha_3`)
    fbreak1 : float
        1st break frequency for the triple broken power law (:math:`f_1`)
    fbreak2 : float
        2nd break frequency for the triple broken power law (:math:`f_2`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        super(TripleBrokenPowerLawModel, self).__init__(**kwargs)

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {
                "omega_ref": None,
                "alpha_1": None,
                "alpha_2": None,
                "alpha_3": None,
                "fbreak1": None,
                "fbreak2": None,
            }
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        frequencies = bline.frequencies
        piecewise = np.piecewise(
            frequencies,
            [
                frequencies <= self.parameters["fbreak1"],
                (frequencies <= self.parameters["fbreak2"])
                & (frequencies > self.parameters["fbreak1"]),
            ],
            [
                lambda frequencies: (frequencies / self.parameters["fbreak1"])
                ** (self.parameters["alpha_1"]),
                lambda frequencies: (frequencies / self.parameters["fbreak1"])
                ** (self.parameters["alpha_2"]),
                lambda frequencies: (
                    self.parameters["fbreak2"] / self.parameters["fbreak1"]
                )
                ** (self.parameters["alpha_2"])
                * (frequencies / self.parameters["fbreak2"])
                ** (self.parameters["alpha_3"]),
            ],
        )
        return self.parameters["omega_ref"] * piecewise


class SmoothBrokenPowerLawModel(GWBModel):
    r"""

    The smooth broken power law is defined as:

    .. math::
        \Omega(f) = \Omega_{\text{ref}}\left(\frac{f}{f_{\text{ref}}}\right) ^{\alpha_1} \left[1+\left(\frac{f}{f_{\text{ref}}}\right)^{\Delta}\right]^{\frac{\alpha_2-\alpha_1}{\Delta}}
    Parameters:
    -----------
    omega_ref : float
        amplitude of signal (:math:`\Omega_{\text{ref}}`)
    Delta : float
        smoothing variable for the smooth broken power law (:math:`\Delta`)
    alpha_1 : float
        low-frequency spectral index of the smooth broken power law (:math:`\alpha_1`)
    alpha_2 : float
        high-frequency spectral index of the smooth broken power law (:math:`\alpha_2`)
    fbreak : float
        break frequency for the smooth broken power law (:math:`f_{\text{ref}}`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        super(SmoothBrokenPowerLawModel, self).__init__(**kwargs)

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {
                "omega_ref": None,
                "fbreak": None,
                "alpha_1": None,
                "alpha_2": None,
                "delta": None,
            }
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        return (
            self.parameters["omega_ref"]
            * (bline.frequencies / self.parameters["fbreak"])
            ** self.parameters["alpha_1"]
            * (
                1
                + (bline.frequencies / self.parameters["fbreak"])
                ** (self.parameters["delta"])
            )
            ** (
                (self.parameters["alpha_2"] - self.parameters["alpha_1"])
                / self.parameters["delta"]
            )
        )


class SchumannModel(GWBModel):
    r"""

    The Schumann model is defined as:


    .. math::
         \Omega(f) = \sum_{ij} \kappa_i \kappa_j \left(\frac{f}{f_{\text{ref}}}\right)^{-\beta_i-\beta_j} M_{ij}(f) \times 10^{-46}
    Parameters:
    -----------
    fref : float
        reference frequency for defining the model (:math:`f_{\text{ref}}`)
    kappa_i : float
        amplitude of coupling function of interferometer i at 10 Hz (:math:`\kappa_i`)
    beta_i : float
        spectral index of coupling function of interferometer i (:math:`\beta_i`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        # Set valid ifos
        # get ifo's associated with this set of
        # baselines.
        valid_ifos = ["H", "L", "V", "K"]
        self.ifos = []
        for bline in kwargs["baselines"]:
            if bline.name[0] not in valid_ifos:
                raise ValueError(
                    "baseline names must be two ifo letters from list of H, L, V, K"
                )
            if bline.name[1] not in valid_ifos:
                raise ValueError(
                    "baseline names must be two ifo letters from list of H, L, V, K"
                )
        # get list of ifos
        for bline in kwargs["baselines"]:
            self.ifos.append(bline.name[0])
            self.ifos.append(bline.name[1])
        super(SchumannModel, self).__init__(**kwargs)
        # set error handling to make sure baselines
        # are named properly
        # make it unique
        # be careful, this doesn't preserve
        # the order!
        self.ifos = list(set(self.ifos))

    @property
    def parameters(self):
        if self._parameters is None:
            schu_params = {}
            for ifo in self.ifos:
                schu_params[f"kappa_{ifo}"] = None
                schu_params[f"beta_{ifo}"] = None
            return schu_params
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        # assumes simple power law model for transfer function
        ifo1 = bline.name[0]
        ifo2 = bline.name[1]
        TF1 = (
            self.parameters[f"kappa_{ifo1}"]
            * (bline.frequencies / 10) ** (-self.parameters[f"beta_{ifo1}"])
            * 1e-23
        )
        TF2 = (
            self.parameters[f"kappa_{ifo2}"]
            * (bline.frequencies / 10) ** (-self.parameters[f"beta_{ifo2}"])
            * 1e-23
        )
        # units of transfer function strain/pT
        return TF1 * TF2 * np.real(bline.M_f)


class TVSPowerLawModel(GWBModel):
    r"""
    The Tensor-Vector-Scalar polarization (T,V,S) power-law model is defined as:

    .. math::

        \Omega(f) = \Omega _T + \Omega _V + \Omega _S

        \Omega _T = \Omega _{{\text{ref}},T} \left( \frac{f}{f_{\text{ref}}}\right)^{\alpha _T}

        \Omega _V = (\gamma _V/\gamma_T)~\Omega _{{\text{ref}},V} \left( \frac{f}{f_{\text{ref}}}\right)^{\alpha _V}

        \Omega _S = (\gamma_S/\gamma_T)~\Omega _{{\text{ref}},S} \left( \frac{f}{f_{\text{ref}}}\right)^{\alpha _S}
    Parameters:
    -----------
    fref : float
        reference frequency for defining the model (:math:`f_{\text{ref}}`)
    omega_ref_pol : float
        amplitude of signal at fref for polarization pol (:math:`\Omega_{\text{ref},\text{pol}}`)
    alpha_pol : float
        spectral index of the power law for polarization pol (:math:`\alpha_{\text{pol}}`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        try:
            fref = kwargs.pop("fref")
        except KeyError:
            raise KeyError("fref must be supplied")
        super(TVSPowerLawModel, self).__init__(**kwargs)
        self.fref = fref

    @property
    def parameters(self):
        if self._parameters is None:
            params = {}
            for pol in self.polarizations:
                params[f"omega_ref_{pol}"] = None
                params[f"alpha_{pol}"] = None
            return params
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        model = np.zeros(bline.frequencies.shape)
        for pol in self.polarizations:
            orf_pol = getattr(bline, f"{pol}_overlap_reduction_function")
            orf_parent = getattr(bline, f"{self.polarizations[0]}_overlap_reduction_function")
            model += (
                orf_pol
                / orf_parent
                * self.parameters[f"omega_ref_{pol}"]
                * (bline.frequencies / self.fref) ** (self.parameters[f"alpha_{pol}"])
            )
        return model


# Parity violation models


class PVPowerLawModel(GWBModel):
    r"""
    The parity violation model can be defined as:

    .. math::
        \Omega(f) = \left(1 + \Pi \frac{\gamma _V}{\gamma _I}\right) \Omega_{\text{ref}} \left( \frac{f}{f_{\text{ref}}} \right)^{\alpha}
    Parameters:
    -----------
    fref : float
        reference frequency for defining the model (:math:`f_{\text{ref}}`)
    omega_ref : float
        amplitude of signal at fref (:math:`\Omega_{\text{ref}}`)
    alpha : float
        spectral index of the power law (:math:`\alpha`)
    Pi : float
        degree of parity violation (:math:`\Pi`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        try:
            fref = kwargs.pop("fref")
        except KeyError:
            raise KeyError("fref must be supplied")
        super(PVPowerLawModel, self).__init__(**kwargs)
        self.fref = fref

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {"omega_ref": None, "alpha": None, "Pi": None}
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        return (
            (
                1
                + self.parameters["Pi"]
                * bline.gamma_v
                / bline.overlap_reduction_function
            )
            * self.parameters["omega_ref"]
            * (bline.frequencies / self.fref) ** (self.parameters["alpha"])
        )


class PVPowerLawModel2(GWBModel):
    r"""
    The parity violation model 2 can be defined as:

    .. math::
         \Omega(f) = \left(1 + f^{\beta} \frac{\gamma_V}{\gamma _I}\right) \Omega_{\text{ref}}\left(\frac{f}{f_{\text{ref}}} \right)^{\alpha}
    Parameters:
    -----------
    fref : float
        reference frequency for defining the model (:math:`f_{\text{ref}}`)
    omega_ref : float
        amplitude of signal at fref (:math:`\Omega_{\text{ref}}`)
    alpha : float
        spectral index of the power law (:math:`\alpha`)
    beta : float
        spectral index of the degree of parity violation (:math:`\beta`)
    frequencies : numpy.ndarray
        array of frequencies at which to evaluate the model
    """

    def __init__(self, **kwargs):
        try:
            fref = kwargs.pop("fref")
        except KeyError:
            raise KeyError("fref must be supplied")
        super(PVPowerLawModel2, self).__init__(**kwargs)
        self.fref = fref

    @property
    def parameters(self):
        if self._parameters is None:
            # include this so that the parent class GWBModel doesn't complain
            return {"omega_ref": None, "alpha": None, "beta": None}
        elif isinstance(self._parameters, dict):
            return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        if parameters is None:
            self._parameters = parameters
        elif isinstance(parameters, dict):
            self._parameters = parameters
        else:
            raise ValueError(f"unexpected type for parameters {type(parameters)}")

    def model_function(self, bline):
        return (
            (
                1
                + ((bline.frequencies) ** (self.parameters["beta"]))
                * bline.gamma_v
                / (bline.overlap_reduction_function)
            )
            * self.parameters["omega_ref"]
            * (bline.frequencies / self.fref) ** (self.parameters["alpha"])
        )



#!/usr/bin/env python

import os

from setuptools import setup


def get_long_description():
    """Finds the README and reads in the description"""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, "README.md")) as f:
        long_description = f.read()
    return long_description


# get version info from __init__.py
def readfile(filename):
    with open(filename) as fp:
        filecontents = fp.read()
    return filecontents


long_description = get_long_description()

setup(
    name="pygwb",
    description="Lighweight python stochastic GWB analysis pipeline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.ligo.org/pygwb/pygwb",
    author="Arianna Renzini, Sylvia Biscoveanu, Shivaraj Khandasamy, Kamiel Janssens, Max Lalleman, Katarina Martinovic, Andrew Matas, Patrick Meyers, Alba Romero, Colm Talbot, Leo Tsukada, Kevin Turbang",
    author_email="arianna.renzini@ligo.org",
    license="MIT",
    packages=["pygwb"],
    package_dir={"pygwb": "pygwb"},
    scripts=['pygwb_pipe/pygwb_pipe','pygwb_pipe/pygwb_combine', 'pygwb_pipe/pygwb_stats', 'pygwb_pipe/pygwb_dag', 'pygwb_pipe/pygwb_pe', 'pygwb_pipe/pygwb_create_isotropic_workflow', 'pygwb_pipe/pygwb_html'],
    install_requires=[
        "numpy",
        "matplotlib",
        "scipy>=1.8.0",
        "bilby>=1.4",
        "gwpy>=3.0.4",
        "astropy==5.2",
        "gwdetchar",
        "gwsumm",
        "pycondor",
        "lalsuite>=7.3",
        "loguru",
        "json5",
        "seaborn",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)

